# Base HQQ/VLLM model defintion
import torch
from torch import float16
from vllm.model_executor.layers.linear import UnquantizedLinearMethod
from ..base import BaseHQQModel, HQQLinear


class HQQLinearMethod(UnquantizedLinearMethod):
    def apply_weights(self, weights, x, bias=None):
        return weights["hqq_module"](x)


class BaseHQQVLLMModel(BaseHQQModel):
    @classmethod
    def quantize_model_single_worker(
        cls,
        model,
        quant_config: dict,
        compute_dtype: torch.dtype = float16,
        device="cuda",
    ):
        # Use the same quantization config for all linear layers. Use None to skip quantizing a specfic layer.
        patch_params = dict([(k, quant_config) for k in cls.get_linear_tags()])

        # We replace the nn.Linear layers with HQQLinear
        def _patch_linear(linear_layer, quant_config):
            if quant_config is None:
                return linear_layer

            hqq_module = HQQLinear(
                linear_layer,
                quant_config,
                compute_dtype=compute_dtype,
                device=device,
                del_orig=False,
            )

            # Clear original params
            del linear_layer.weight
            linear_layer.bias = None  # bias is inside hqq_module

            # Set HQQ params
            linear_layer.linear_weights = {"hqq_module": hqq_module}
            linear_layer.linear_method = HQQLinearMethod()

            torch.cuda.empty_cache()

            return linear_layer

        cls.patch_model(
            model,
            lambda layer: layer.to(device=device, dtype=compute_dtype),
            _patch_linear,
            patch_params,
        )

    @classmethod
    def quantize_model(
        cls, model, quant_config, compute_dtype: torch.dtype = float16, device="cuda"
    ):
        workers = model.llm_engine.workers
        for i in range(len(workers)):
            cls.quantize_model_single_worker(
                workers[i].model,
                quant_config=quant_config,
                compute_dtype=compute_dtype,
                device=device,
            )

    # Save model architecture
    @classmethod
    def cache_model(cls, model, save_dir):
        model_0 = model.llm_engine.workers[0].model
        model_0.config.save_pretrained(save_dir)

    @classmethod
    def cache_model_single_worker(cls, model_0, save_dir):
        model_0.config.save_pretrained(save_dir)

    @classmethod
    def serialize_weights(cls, model, verbose):
        weights = {}
        ignore_keys = cls.get_ignore_layers(model)
        for name, module in model.named_modules():
            if name in ignore_keys:
                continue
            try:
                state_dict = module.state_dict()
                if len(state_dict) > 0:
                    weights[name] = dict(state_dict)
                else:
                    # Quantized linear layers in a VLLM model
                    if hasattr(module, "linear_weights"):
                        if "hqq_module" in module.linear_weights:
                            weights[name] = module.linear_weights[
                                "hqq_module"
                            ].state_dict()

            except Exception:
                if verbose:
                    print("Skipping", name)

        return weights

    @classmethod
    def save_quantized(cls, model, save_dir: str, verbose: bool = False):
        model_0 = model.llm_engine.workers[0].model

        # Cache model
        cls.cache_model_single_worker(model_0, save_dir)

        # Serialization
        weights = cls.serialize_weights(model_0, verbose=verbose)

        # Save
        cls.save_weights(weights, save_dir)

    #################################################
    @classmethod
    def from_quantized_single_worker(
        cls,
        save_dir_or_hub: str,
        cache_dir: str = "",
        compute_dtype: torch.dtype = float16,
        device="cuda:0",
    ):
        # Get directory path
        save_dir = cls.try_snapshot_download(save_dir_or_hub, cache_dir)

        # Load model from config
        model = cls.create_model(save_dir)

        # Name the layers
        cls.autoname_modules(model)

        # Load weights
        try:
            weights = cls.load_weights(save_dir)
        except Exception as error:
            print("Failed to load the weights", error)
            return

        # load_state_dict() doesn't work with modules initialized with init_empty_weights(), so we need to do this manually
        @torch.no_grad()
        def _load_module(module, params=None):
            if module.name not in weights:
                return module.to(device=device, dtype=compute_dtype)

            state_dict = weights[module.name]
            if ("W_q" in state_dict) and ("meta" in state_dict):
                hqq_module = HQQLinear(
                    linear_layer=None,
                    quant_config=None,
                    compute_dtype=compute_dtype,
                    device=device,
                )
                hqq_module.load_state_dict(state_dict)

                # Clear original params
                del module.weight
                module.bias = None  # bias is inside hqq_module

                # Set HQQ params
                module.linear_weights = {"hqq_module": hqq_module}
                module.linear_method = HQQLinearMethod()

                torch.cuda.empty_cache()

            else:
                for key in state_dict:
                    setattr(
                        module,
                        key,
                        torch.nn.Parameter(
                            state_dict[key].to(device), requires_grad=False
                        ),
                    )

            return module

        # Load modules
        cls.patch_model(
            model,
            _load_module,
            _load_module,
            dict([(k, None) for k in cls.get_linear_tags()]),
        )
        # Load other weights that are not part of any module
        cls.post_module_load(model, weights)

        return model



# Written by Dr. Hicham Badri @Mobius Labs GmbH - 2023
#####################################################
from ..base import BasePatch
from .base import BaseHQQHFModel
from tqdm import tqdm
import torch

# Experimental - This is the optimized version of microsoft's Phi model which contains merged QKV layers and other things.
# Unfortunately this cannot be used automatically because the base class PhiForCausalLM has been modified. you need to use it manually by calling PhiHQQ.fct(model, args)


# Patch functions
class PhiPatch(BasePatch):
    # These tags are used to specify the parameters of each layer type. For example, if you want to give different quantization parameters to different layers
    @classmethod
    def get_linear_tags(cls):
        return ["mixer.Wqkv", "mixer.out_proj", "mlp.fc1", "mlp.fc2"]

    @classmethod
    def patch_nonlinearlayers(cls, model, patch_fct, verbose=True):
        # Remove dropout
        model.transformer.embd.drop = torch.nn.Identity()

        base_model = model.transformer
        model.lm_head = patch_fct(model.lm_head)
        base_model.embd = patch_fct(base_model.embd)

        layers = base_model.h
        for i in tqdm(range(len(layers)), disable=not verbose):
            layers[i].ln = patch_fct(layers[i].ln)
            layers[i].mixer.rotary_emb = patch_fct(layers[i].mixer.rotary_emb)
            layers[i].mlp.act = patch_fct(layers[i].mlp.act)

    @classmethod
    def patch_linearlayers(cls, model, patch_fct, patch_params, verbose=True):
        base_model = model.transformer
        layers = base_model.h
        for i in tqdm(range(len(layers)), disable=not verbose):
            layers[i].mixer.Wqkv = patch_fct(
                layers[i].mixer.Wqkv, patch_params["mixer.Wqkv"]
            )
            layers[i].mixer.out_proj = patch_fct(
                layers[i].mixer.out_proj, patch_params["mixer.out_proj"]
            )
            layers[i].mlp.fc1 = patch_fct(layers[i].mlp.fc1, patch_params["mlp.fc1"])
            layers[i].mlp.fc2 = patch_fct(layers[i].mlp.fc2, patch_params["mlp.fc2"])

            # Remove dropout
            layers[i].resid_dropout = torch.nn.Identity()
            layers[i].mixer.inner_attn.drop = torch.nn.Identity()
            layers[i].mixer.inner_cross_attn.drop = torch.nn.Identity()


class PhiHQQ(PhiPatch, BaseHQQHFModel):
    pass



# Written by Dr. Hicham Badri @Mobius Labs GmbH - 2023
#####################################################

# Import VLLM architectures here with the dummy_load trick
import torch
from torch import float16
import transformers
from ..models.vllm.llama import LlamaForCausalLM, LlamaHQQ
from ..models.base import BaseHQQModel
from .base import HQQWrapper
from termcolor import colored

# Set them in the model registry
import vllm

vllm.model_executor.model_loader._MODEL_REGISTRY["LlamaForCausalLM"] = LlamaForCausalLM

_HQQ_REGISTRY = {}
_HQQ_REGISTRY["LlamaForCausalLM"] = LlamaHQQ

# VLLM requires model input as string, so this is the fallback for a dummy init used to load the model from quantized weights + load the tokenizer
_ARCH_TO_DEFAULT = {}
_ARCH_TO_DEFAULT["LlamaForCausalLM"] = "meta-llama/Llama-2-7b-chat-hf"

_Parent = vllm.entrypoints.llm.LLM


# Similar to the from vllm import LLM class, with some extra parameters and only loads a dummy model
class HQQLLM(_Parent, HQQWrapper):
    _HQQ_REGISTRY = _HQQ_REGISTRY
    INIT_CACHE_MEM = 1.0  # Default cache mem in GB. This is extra-memory, but you need it otherwise runtime is slower

    def __init__(self, *args, **kwargs):
        self._validate_params(kwargs)

        if "force_skip" in kwargs:
            force_skip = kwargs["force_skip"]
            del kwargs["force_skip"]
        else:
            force_skip = False

        # This will init a dummy model
        super().__init__(*args, **kwargs)

        cache_dir = kwargs["cache_dir"] if ("cache_dir" in kwargs) else ""
        model_id = kwargs["model"]

        # Here we load the model on CPU so that we can quantize it while avoiding extra GPU cost
        self.arch_key = self.llm_engine.model_config.hf_config.architectures[0]
        model_config = self.llm_engine.model_config.hf_config
        workers = self.llm_engine.workers

        if force_skip is False:
            for i in range(len(workers)):
                workers[i].model.__init__(model_config, dummy_load=False)
                workers[i].model.load_weights(
                    model_name_or_path=model_id, cache_dir=cache_dir
                )

        self.hqq_quantized = False

    def _validate_params(self, kwargs: dict):
        if "gpu_memory_utilization" not in kwargs:
            total_gpu_mem = torch.cuda.get_device_properties(0).total_memory / 1e9  # GB
            kwargs["gpu_memory_utilization"] = HQQLLM.INIT_CACHE_MEM / total_gpu_mem

    # In case the user wants to use the fp16 model and skip quantization
    def cuda(self):
        self._check_if_already_quantized(self)
        workers = self.llm_engine.workers
        for i in range(len(workers)):
            workers[i].model = workers[i].model.half().cuda(i)
        return self

    def quantize_model(
        self, quant_config: dict, compute_dtype: torch.dtype = float16, device="cuda"
    ):
        return self.quantize_model_(
            model=self,
            quant_config=quant_config,
            compute_dtype=compute_dtype,
            device=device,
        )

    def save_quantized(self, save_dir: str):
        return self.save_quantized_(model=self, save_dir=save_dir)

    @classmethod
    def _get_arch_key_from_save_dir(cls, save_dir: str):
        config = transformers.AutoConfig.from_pretrained(save_dir)
        return config.architectures[0]

    # This requires custom loading because VLLM requires a str input model to initalize the workers
    @classmethod
    def from_quantized(
        cls,
        save_dir_or_hub: str,
        compute_dtype: torch.dtype = float16,
        cache_dir: str = "",
        tensor_parallel_size: int = 1,
    ):
        assert tensor_parallel_size == 1, "Only single GPU is supported."
        # Both local and hub-support
        save_dir = BaseHQQModel.try_snapshot_download(save_dir_or_hub)
        arch_key = cls._get_arch_key_from_save_dir(save_dir)
        cls._check_arch_support(arch_key)

        # Trick to initialize the tokenizer and a dummy model inside a VLLM instance
        instance = cls(
            model=_ARCH_TO_DEFAULT[arch_key], force_skip=True, tensor_parallel_size=1
        )
        workers = instance.llm_engine.workers
        for i in range(len(workers)):
            workers[i].model = cls._get_hqq_class(
                arch_key
            ).from_quantized_single_worker(
                save_dir_or_hub=save_dir_or_hub,
                compute_dtype=compute_dtype,
                cache_dir=cache_dir,
                device="cuda:" + str(i),
            )

        cls._make_quantizable(instance, quantized=True)
        return instance


# From https://github.com/langchain-ai/langchain/blob/master/libs/langchain/langchain/llms/vllm.py


try:
    from langchain.llms import VLLM as LangchainVLLMBase
    from langchain_core.pydantic_v1 import root_validator

    class LangchainVLLM(LangchainVLLMBase):
        def set(self, model):
            self.client = model
            return self

        @root_validator()
        def validate_environment(cls, values: dict) -> dict:
            return values
except Exception:
    LangchainVLLM = None
    print(
        colored(
            'Langchain not installed. You can install it via "pip install langchain"',
            "cyan",
        )
    )



# Written by Dr. Hicham Badri @Mobius Labs GmbH - 2023
#####################################################
import torch
import numpy as np
from torch import float32, float16, Tensor
from functools import partial
from typing import Union


# re-estimate the scale based on the inverse median: Only tested with axis==0
def update_scale_inverse_median(
    W_f: Tensor, scale: Tensor, zero: Tensor, axis: int, min_max: list
) -> tuple:
    scale_rng = 2e4
    z_val = 1e-4
    delta = 1e-2

    W_q = torch.round(W_f * scale + zero).clamp(min_max[0], min_max[1])

    # Correct zero to avoid W_q==zero
    zero_c = zero.clone()
    zero_c_indx = torch.sum(1.0 * ((W_q - zero) == 0), axis=axis, keepdim=True) > 0
    zero_c[zero_c_indx] = zero_c[zero_c_indx] + delta

    # Build scale tensor
    W_f_c = W_f.clone()
    W_f_c_mask = torch.abs(W_f_c) < z_val
    W_f_c[W_f_c_mask] = z_val

    scale_tensor = (W_q - zero_c).float() / W_f_c.float()
    # W_r               = (W_q - zero_c)/scale_tensor

    # Normalize scale_tensor
    scale_b = torch.median(scale_tensor, axis=axis, keepdim=True)[0]
    scale_b = scale_b.clamp(min=-scale_rng, max=scale_rng).half()

    # Mix with older scale
    W_r = (W_q - zero_c) / scale_b
    err_b = torch.abs(W_f - W_r).mean(axis=axis, keepdim=True)

    W_r = (W_q - zero_c) / scale
    err_a = torch.abs(W_f - W_r).mean(axis=axis, keepdim=True)

    mask = (err_b < err_a).half()
    scale_b = mask * scale_b + (1 - mask) * scale

    # W_r   = (W_q - zero_c)/scale_b
    return scale_b, zero_c


# Greedy local search: Only tested with axis==0
def update_scale_grid_search(
    W_f: Tensor, scale: Tensor, zero: Tensor, axis: int, min_max: list, N: int = 128 + 1
) -> Tensor:
    # Make sure it's an odd number so that the original scale is included
    assert N % 2 == 1, "Please check whether N: odd number"
    rng_dump = 0.05  # 0.05 / 1.
    z_val = 2e-4

    device = scale.device
    dtype = scale.dtype
    ###############################
    W_q = torch.round(W_f * scale + zero).clamp(min_max[0], min_max[1])
    n_clusters = max(W_q.shape[0], W_q.shape[1])
    rng = torch.abs(scale).mean() * rng_dump if (rng_dump < 1.0) else rng_dump

    scale_shifted = (
        torch.linspace(-rng, rng, N)[:, None]
        .to(dtype=dtype, device=device)
        .repeat(1, n_clusters)
        + scale
    )

    # Safe inverse
    scale_shifted[
        torch.logical_and(scale_shifted >= 0, torch.abs(scale_shifted) <= z_val)
    ] = z_val
    scale_shifted[
        torch.logical_and(scale_shifted < 0, torch.abs(scale_shifted) <= z_val)
    ] = -z_val

    err = torch.empty([N, n_clusters], dtype=dtype, device=device)
    for i in range(N):
        W_r = (W_q - zero) / scale_shifted[i][None, :]
        err[i] = torch.abs(W_f - W_r).mean(axis=axis, keepdim=True)

    ind_r = torch.argmin(err, axis=axis).to(torch.int32)
    ind_c = torch.arange(len(ind_r), dtype=torch.int32, device=device)
    scale_b = scale_shifted[ind_r, ind_c]

    return scale_b


# Shrinking operator
def shrink_lp_op(x: Tensor, beta: float, lp_norm: float) -> Tensor:
    if lp_norm == 1:
        return torch.sign(x) * torch.nn.functional.relu(torch.abs(x) - 1.0 / beta)
    else:
        return torch.sign(x) * torch.nn.functional.relu(
            torch.abs(x) - (1.0 / beta) * torch.pow(torch.abs(x), lp_norm - 1)
        )


# Proximal solver || W - dequantize(quantize(W))||_p^p - Experimental
@torch.inference_mode()
def optimize_weights_proximal_v2(
    tensor: Tensor,
    scale: Tensor,
    zero: Tensor,
    min_max: list,
    axis: int = 0,
    device: Union[str, None] = None,
    dtype: Union[torch.dtype, None] = None,
    opt_params: dict = {
        "lp_norm": 0.7,
        "beta": 1e1,
        "kappa": 1.01,
        "iters": 20,
        "tol": 0.0,
        "early_stop": True,
        "scale_gridsearch": False,
    },
    verbose: bool = False,
) -> tuple:
    # Params
    lp_norm = max(opt_params["lp_norm"], 0.1)
    beta = opt_params["beta"]
    kappa = opt_params["kappa"]
    iters = opt_params["iters"]
    early_stop = opt_params["early_stop"]
    tol = opt_params["tol"]

    # Check
    assert lp_norm <= 1.0, "lp_norm should be <=1"
    assert beta > 0.0, "beta should be > 0"
    assert kappa > 1.0, "kappa should be > 1"
    assert iters > 1, "iters should be > 1"

    # Cast/device
    if device is None:
        device = tensor.device
    else:
        device = torch.device(device)

    if dtype is None:
        dtype = float16 if (device.type == "cuda") else float32

    W_f = tensor.to(device=device, dtype=dtype)
    scale = scale.to(device=device, dtype=dtype)
    zero = zero.to(device=device, dtype=dtype)

    # Update scale: works slightly better. Tested on Llama2 only
    if opt_params["scale_gridsearch"]:
        scale = update_scale_grid_search(W_f, scale, zero, axis, min_max)

    # Optimize for zero-point
    best_error = 1e4
    scale_prev, zero_prev = scale.clone(), zero.clone()
    for i in range(iters):
        W_q = torch.round(W_f * scale + zero).clamp(min_max[0], min_max[1])
        W_r = (W_q - zero) / scale

        # current_error = float(torch.pow(torch.abs(W_f - W_r), max(0.80, lp_norm)).mean())
        current_error = float(torch.abs(W_f - W_r).mean())

        if verbose:
            print(i, np.round(current_error, 6))

        if early_stop:
            if best_error - current_error > tol:
                best_error = current_error
                scale_prev, zero_prev = scale.clone(), zero.clone()
            else:
                scale, zero = scale_prev.clone(), zero_prev.clone()
                break

        W_e = shrink_lp_op(W_f - W_r, beta, lp_norm)
        zero = torch.mean(W_q - (W_f - W_e) * scale, axis=axis, keepdim=True)
        beta *= kappa

    # Clean-up
    scale = scale.to(tensor.device)
    zero = zero.to(tensor.device)
    del W_f, W_q, W_r, W_e, scale_prev, zero_prev
    torch.cuda.empty_cache()

    W_q = torch.round(tensor * scale + zero).clamp(min_max[0], min_max[1])

    return W_q, scale, zero


# Proximal solver || W - dequantize(quantize(W))||_p^p
@torch.inference_mode()
def optimize_weights_proximal_legacy(
    tensor: Tensor,
    scale: Tensor,
    zero: Tensor,
    min_max: list,
    axis: int = 0,
    device: Union[str, None] = None,
    opt_params: dict = {"lp_norm": 0.7, "beta": 1e1, "kappa": 1.01, "iters": 20},
    verbose: bool = False,
) -> tuple:
    lp_norm, beta, kappa, iters = (
        opt_params["lp_norm"],
        opt_params["beta"],
        opt_params["kappa"],
        opt_params["iters"],
    )

    if device is None:
        device = tensor.device
    else:
        device = torch.device(device)

    dtype = float16 if (device.type == "cuda") else float32
    W_f = tensor.to(dtype=dtype, device=device)
    scale = scale.to(dtype=dtype, device=device)
    zero = zero.to(dtype=dtype, device=device)

    best_error = 1e4
    for i in range(iters):
        W_q = torch.round(W_f * scale + zero).clamp(min_max[0], min_max[1])
        W_r = (W_q - zero) / scale
        W_e = shrink_lp_op(W_f - W_r, beta, lp_norm)
        zero = torch.mean(W_q - (W_f - W_e) * scale, axis=axis, keepdim=True)
        beta *= kappa

        current_error = float(torch.abs(W_f - W_r).mean())
        if verbose:
            print(i, np.round(current_error, 6))
        if current_error < best_error:
            best_error = current_error
        else:
            break

    scale = scale.to(tensor.device)
    zero = zero.to(tensor.device)
    del W_f, W_q, W_r, W_e
    torch.cuda.empty_cache()

    W_q = torch.round(tensor * scale + zero).clamp(min_max[0], min_max[1])
    return W_q, scale, zero


# Default: fast with early stopping
optimize_weights_proximal = optimize_weights_proximal_legacy

# Slower, better quality: no early stoppping, more iterations
optimize_weights_proximal_slow = partial(
    optimize_weights_proximal_v2,
    dtype=torch.float32,
    opt_params={
        "lp_norm": 0.7,
        "beta": 1e1,
        "kappa": 1.01,
        "iters": 100,
        "tol": 0.0,
        "early_stop": False,
        "scale_gridsearch": False,
    },
)

##############################################################################################################
# L1 with SGD optimizer: supports scale and W_q updates. L{p<1} fails with SGD


class LinearSchedulerWithWarmStart(torch.optim.lr_scheduler._LRScheduler):
    def __init__(self, optimizer, lr_start, lr_end, iters, warm_start=0, last_epoch=-1):
        iters_wrm = max(0, int(iters * warm_start))
        self.lr_wrm = (
            np.linspace(lr_end, lr_start, iters_wrm)
            if (iters_wrm > 0)
            else np.array([])
        )
        self.lr_mid = np.linspace(lr_start, lr_end, iters - iters_wrm)
        self.lr_sch = np.concatenate([self.lr_wrm, self.lr_mid])
        self.idx = 0
        super(LinearSchedulerWithWarmStart, self).__init__(optimizer, last_epoch)

    def get_lr(self):
        self.idx = min(self.idx, len(self.lr_sch) - 1)
        out = [self.lr_sch[self.idx] for base_lr in self.base_lrs]
        self.idx += 1
        return out


# SGD solver  || W - dequantize(quantize(W))||_1 (p=1 only, with additional fake inputs x)
def optimize_weights_autograd(
    tensor: Tensor,
    scale: Tensor,
    zero: Tensor,
    min_max: list,
    axis: int = 0,
    device: Union[str, None] = None,
    dtype: Union[torch.dtype, None] = float32,
    opt_params: dict = {
        "lr": 2e-3,
        "iters": 2500,
        "lr_schedule": False,
        "update_Wq": False,
        "use_fake_data": False,
    },
    data_params: dict = {"normalize": False, "data_rng": 10.0, "data_ctx": 32},
    compile: bool = True,
    verbose: bool = False,
) -> tuple:
    ref_device = scale.device
    ref_dtype = scale.dtype

    if device is None:
        device = tensor.device
    else:
        device = torch.device(device)

    W_f = tensor.to(dtype=dtype, device=device)

    params = {}
    params["scale"] = torch.nn.Parameter(
        scale.to(dtype=dtype, device=device), requires_grad=True
    )
    params["zero"] = torch.nn.Parameter(
        zero.to(dtype=dtype, device=device), requires_grad=True
    )

    if opt_params["update_Wq"]:
        with torch.no_grad():
            params["W_q"] = torch.round(W_f * params["scale"] + params["zero"]).clamp(
                min_max[0], min_max[1]
            )
        params["W_q"] = torch.nn.Parameter(params["W_q"], requires_grad=True)

    optimizer = torch.optim.AdamW(
        [params[k] for k in params],
        lr=opt_params["lr"],
        betas=(0.9, 0.99),
        eps=1e-06,
        weight_decay=0.0,
    )

    if opt_params["lr_schedule"]:
        scheduler = LinearSchedulerWithWarmStart(
            optimizer,
            lr_start=opt_params["lr"],
            lr_end=1e-6,
            iters=opt_params["iters"],
            warm_start=0,
        )
    else:
        scheduler = None

    with torch.no_grad():
        if data_params["normalize"]:
            scale_loss = 1.0 / (tensor.abs().mean() + 1e-4)
        else:
            scale_loss = 1.0

    def _loss_fct(output, target):
        return torch.mean(torch.abs(scale_loss * (target - output)))  # L1

    def _fake_quant_fixed_Wq(W_f):
        # Quantize
        W_q = torch.round(W_f * params["scale"] + params["zero"]).clamp(
            min_max[0], min_max[1]
        )
        # Dequantize
        W_r = (W_q - params["zero"]) / params["scale"]
        return W_r

    def _fake_quant_update_Wq(W_f):
        # Quantize
        W_q = torch.round(params["W_q"]).clamp(min_max[0], min_max[1])
        # Dequantize
        W_r = (W_q - params["zero"]) / params["scale"]
        return W_r

    if opt_params["update_Wq"]:
        _fake_quant = _fake_quant_update_Wq
    else:
        _fake_quant = _fake_quant_fixed_Wq

    def _fake_quant_loss(W_f):
        return _loss_fct(_fake_quant(W_f), W_f)

    def _fake_quant_loss_with_fake_data(W_f):
        x = (
            torch.rand(
                [data_params["data_ctx"], W_f.shape[1]], device=device, dtype=dtype
            )
            - 0.5
        ) * 2 ** data_params["data_rng"]
        y_ref = torch.matmul(x, W_f.T)
        y_pred = torch.matmul(x, _fake_quant(W_f).T)
        return _loss_fct(y_pred, y_ref)

    if opt_params["use_fake_data"]:
        _fake_quant_loss = _fake_quant_loss_with_fake_data
    else:
        _fake_quant_loss = _fake_quant_loss

    if compile:
        _fake_quant_loss = torch.compile(_fake_quant_loss)

    def _step(W_f):
        optimizer.zero_grad()
        loss = _fake_quant_loss(W_f)
        loss.backward()
        optimizer.step()
        if scheduler is not None:
            scheduler.step()
        return np.round(loss.item(), 10)

    with torch.no_grad():
        _init_loss = _fake_quant_loss(W_f).item()

    for i in range(opt_params["iters"]):
        loss_out = _step(W_f)
        if verbose and (i % 100) == 0:
            print(i, loss_out)

    with torch.no_grad():
        _final_loss = _fake_quant_loss(W_f).item()

    if _final_loss < _init_loss:
        for k in params:
            params[k] = params[k].data.detach()
    else:
        if verbose:
            print("optimization failed...")
        params = {"scale": scale, "zero": zero}

    scale = params["scale"].to(device=ref_device, dtype=ref_dtype)
    zero = params["zero"].to(device=ref_device, dtype=ref_dtype)

    if "W_q" in params:
        W_q = params["W_q"].to(device=ref_device, dtype=ref_dtype)
    else:
        W_q = (
            torch.round(tensor * scale + zero)
            .clamp(min_max[0], min_max[1])
            .to(device=ref_device, dtype=ref_dtype)
        )

    del W_f
    torch.cuda.empty_cache()
    return W_q, scale, zero


optimize_weights_autograd_main = partial(
    optimize_weights_autograd,
    dtype=torch.float32,
    opt_params={
        "lr": 2e-3,
        "iters": 1000,
        "lr_schedule": True,
        "update_Wq": True,
        "use_fake_data": False,
    },
    verbose=False,
    compile=True,
)

optimize_weights_autograd_fakedata = partial(
    optimize_weights_autograd,
    dtype=torch.float32,
    opt_params={
        "lr": 2e-3,
        "iters": 1000,
        "lr_schedule": True,
        "update_Wq": True,
        "use_fake_data": True,
    },
    data_params={"normalize": False, "data_rng": 10.0, "data_ctx": 32},
    verbose=False,
    compile=True,
)



# Written by Dr. Hicham Badri @Mobius Labs GmbH - 2023
#####################################################
import torch
from torch import float16, bfloat16, float32
from torch import Tensor, nn
import numpy as np
from .quantize import HQQLinear, Quantizer
from .utils import cleanup
from typing import Union


# Return trainable weight matrix
def _get_dense_param(
    in_features: int,
    out_features: int,
    device="cuda",
    trainable: bool = True,
    dtype: torch.dtype = float32,
) -> Tensor:
    W = (
        nn.Linear(in_features, out_features, bias=None)
        .weight.data.t()
        .to(dtype)
        .to(device)
        .contiguous()
    )
    return nn.Parameter(W, requires_grad=trainable)


# Applies LoRA to a linear module (include HQQLinear)
class HQQLinearLoRA(nn.Module):
    def __init__(self, linear_layer: nn.Module, peft_config: dict):
        super().__init__()

        # Device
        self.device = (
            linear_layer.device
            if hasattr(linear_layer, "device")
            else next(linear_layer.parameters()).device
        )
        self.train_dtype = (
            peft_config["train_dtype"]
            if ("train_dtype" in peft_config)
            else torch.float
        )

        # Linear layer
        self.linear_layer = linear_layer
        self.in_features = linear_layer.in_features
        self.out_features = linear_layer.out_features

        # Bias
        self.bias = None if (linear_layer.bias is None) else linear_layer.bias.clone()
        self.linear_layer.bias = None
        self.train_bias = (
            peft_config["train_bias"] if ("train_bias" in peft_config) else False
        )

        if self.bias is not None:
            self.bias = nn.Parameter(self.bias, requires_grad=self.train_bias)
            if self.train_bias:
                self.bias.data = self.bias.data.to(self.train_dtype)

        if (self.bias is None) and self.train_bias:
            self.bias = nn.Parameter(
                torch.zeros(
                    (self.out_features,), device=self.device, dtype=self.train_dtype
                ),
                requires_grad=True,
            )

        # Dropout
        if "dropout" in peft_config:
            self.peft_drop = (
                nn.Dropout(p=peft_config["dropout"])
                if (peft_config["dropout"] > 0.0)
                else nn.Identity()
            )
        else:
            self.peft_drop = nn.Identity()

        # LoRA A/B
        self.peft_config = peft_config
        self.lora_alpha = peft_config["lora_alpha"]
        self.r = peft_config["r"]
        self.scaling = self.lora_alpha / self.r

        self.lora_A = _get_dense_param(
            self.in_features,
            self.r,
            device=self.device,
            trainable=True,
            dtype=self.train_dtype,
        )

        self.lora_B = _get_dense_param(
            self.r,
            self.out_features,
            device=self.device,
            trainable=True,
            dtype=self.train_dtype,
        )

        # LoRA weights init
        if "lora_init" in peft_config:
            # Set lora init
            assert (
                peft_config["lora_init"]["lora_A"].shape[0],
                peft_config["lora_init"]["lora_B"].shape[1],
            ) == (self.in_features, self.out_features), (
                "Invalid init LoRA weight shapes. Expected: lora_A: "
                + str(self.in_features)
                + " x r , lora_B: r x "
                + str(self.out_features)
                + ")"
            )
            self.lora_A.data = peft_config["lora_init"]["lora_A"].to(
                device=self.device, dtype=self.train_dtype
            )
            self.lora_B.data = peft_config["lora_init"]["lora_B"].to(
                device=self.device, dtype=self.train_dtype
            )
        else:
            # Init weights, as as the original LoRA implementation
            nn.init.kaiming_uniform_(self.lora_A, a=np.sqrt(5))
            nn.init.zeros_(self.lora_B)

    def forward_lora(self, x: Tensor) -> Tensor:  # output is self.train_dtype
        return (
            torch.matmul(
                torch.matmul(self.peft_drop(x.to(self.lora_A.dtype)), self.lora_A),
                self.lora_B,
            )
            * self.scaling
        )

    def forward(self, x: Tensor) -> Tensor:
        x_dtype = x.dtype

        # Forward with the base linear layer
        out = self.linear_layer(x)

        # LoRA
        if self.train_bias:
            out += (self.forward_lora(x) + self.bias).to(x_dtype)

        else:
            out += self.forward_lora(x).to(x_dtype)
            if self.bias is not None:
                out += self.bias

        return out

    def merge_and_quantize(self, quant_config: dict) -> nn.Module:
        # not easy to figure out the type for "any" potential linear layer (nn.Linear, HQQLinear, etc.)
        for _dtype in [float16, bfloat16, float32]:
            try:
                # Get initial weights
                W = self.linear_layer(
                    torch.eye(self.in_features, device=self.device, dtype=_dtype)
                ).t()  # == self.linear_layer.dequantize()
            except Exception:
                pass

        # Merge weights
        W += (
            (torch.matmul(self.lora_A.data, self.lora_B.data) * self.scaling)
            .t()
            .to(W.dtype)
        )

        # New HQQ layer
        new_hqq_layer = HQQLinear(None, quant_config)
        new_hqq_layer.bias = None if (self.bias is None) else self.bias.clone()
        new_hqq_layer.quantize(W, **quant_config)

        return new_hqq_layer

    def cast(self, dtype: torch.dtype = float16):
        self.lora_A.data = self.lora_A.data.to(dtype)
        self.lora_B.data = self.lora_B.data.to(dtype)
        if self.bias is not None:
            self.bias.data = self.bias.data.to(dtype)
        if isinstance(self.scaling, nn.Parameter):
            self.scaling.data = self.scaling.data.to(dtype)
        if isinstance(self.scaling, Tensor):
            self.scaling = self.scaling.to(dtype)
        return self

    def state_dict(self):
        return {
            "lora_A": self.lora_A.data,
            "lora_B": self.lora_B.data,
            "scaling": self.scaling,
            "bias": self.bias,
        }

    def load_state_dict(self, state_dict):
        self.lora_A.data = state_dict["lora_A"].data.to(self.device)
        self.lora_B.data = state_dict["lora_B"].data.to(self.device)

        if state_dict["bias"] is not None:
            self.bias.data = state_dict["bias"].data.to(self.device)

        # Handle different use-cases of scaling
        if isinstance(state_dict["scaling"], (int, float)):
            self.scaling = state_dict["scaling"]
        if isinstance(state_dict["scaling"], nn.Parameter):
            self.scaling.data = state_dict["scaling"].data.to(self.device)
        if isinstance(state_dict["scaling"], Tensor):
            self.scaling = state_dict["scaling"].to(self.device)


# LoRA with fake quantization
class HQQLinearLoRAWithFakeQuant(HQQLinearLoRA):
    def __init__(self, linear_layer: nn.Module, peft_config: dict):
        super(HQQLinearLoRAWithFakeQuant, self).__init__(linear_layer, peft_config)
        self.quant_param = peft_config["quant_param"]

    # @torch.no_grad()
    # @torch.compile()
    def fake_quant(self, weight):
        if self.quant_param:
            W_q, meta = Quantizer.quantize(
                weight, **self.quant_param, bitpack=False
            )  # todo: clone() tensor
            weight_est = Quantizer.dequantize(W_q, meta)
        else:
            weight_est = weight
        return weight_est

    def forward(self, x: Tensor) -> Tensor:
        x_dtype = x.dtype

        # Get initial weights
        W = self.linear_layer(
            torch.eye(self.in_features, device=self.device, dtype=x_dtype)
        ).t()  # == self.linear_layer.dequantize()

        # Merge weights
        W += (torch.matmul(self.lora_A, self.lora_B) * self.scaling).t().to(W.dtype)

        # Fake quant
        W = self.fake_quant(W).to(x_dtype)

        # Matmul
        out = torch.matmul(x, W.t())

        # Bias
        if self.bias is not None:
            out += self.bias

        out = out.to(x_dtype)

        return out


# Experimental
class HQQLinearGroupedProj(nn.Module):
    def __init__(self, linear_layer: nn.Module, peft_config: dict):
        super().__init__()

        # Device
        self.device = (
            linear_layer.device
            if hasattr(linear_layer, "device")
            else next(linear_layer.parameters()).device
        )
        self.train_dtype = (
            peft_config["train_dtype"]
            if ("train_dtype" in peft_config)
            else torch.float
        )

        # Linear layer
        self.linear_layer = linear_layer
        self.in_features = linear_layer.in_features
        self.out_features = linear_layer.out_features
        self.bias = None if (linear_layer.bias is None) else linear_layer.bias.clone()

        # Turn-off bias in the linear layer
        self.linear_layer.bias = None

        # Group proj
        self.peft_config = peft_config
        self.proj_size = peft_config["proj_size"]
        self.proj_num = peft_config["proj_num"]
        self.proj = nn.Parameter(
            torch.stack(
                [torch.eye(self.proj_size, dtype=self.train_dtype, device=self.device)]
                * self.proj_num
            )
        )
        if peft_config["zero_trainable"]:
            self.linear_layer.meta["zero"] = nn.Parameter(
                self.linear_layer.meta["zero"].to(self.train_dtype), requires_grad=True
            )

    @torch.compile()
    def forward(self, x: Tensor) -> Tensor:
        x_dtype = x.dtype

        # Forward with base linear
        with torch.no_grad():
            W = self.linear_layer.dequantize().clone()
        # W = self.linear_layer(torch.eye(self.in_features, device=self.device, dtype=x_dtype)).t()
        shape = W.shape

        # Grouped proj
        proj_b, gs = self.proj.shape[0], self.proj.shape[1]
        W = (
            torch.matmul(self.proj, W.reshape((proj_b, gs, -1)).to(self.proj.dtype))
            .to(x_dtype)
            .reshape(shape)
        )

        # Matmul
        out = torch.matmul(x, W.t())

        # Bias
        if self.bias is not None:
            out += self.bias

        out = out.to(x_dtype)

        return out

    def cast(self, dtype: torch.dtype = float16):
        self.proj.data = self.proj.data.to(dtype)
        self.linear_layer.meta["zero"] = self.linear_layer.meta["zero"].to(dtype)
        if self.bias is not None:
            if self.bias.requires_grad:
                self.bias.data = self.bias.data.to(dtype)
            else:
                self.bias = self.bias.to(dtype)
        return self

    def state_dict(self):
        return {
            "proj": self.proj.data,
            "bias": self.bias,
            "peft_config": self.peft_config,
        }

    def load_state_dict(self, state_dict):
        self.proj.data = state_dict["proj"].data.to(self.device)
        self.bias = state_dict["bias"] if ("bias" in state_dict) else None
        self.bias = self.bias.to(self.device) if (self.bias is not None) else None
        self.peft_config = state_dict["peft_config"]


_HQQ_LORA_CLASSES = [HQQLinearLoRA, HQQLinearLoRAWithFakeQuant, HQQLinearGroupedProj]
_HQQ_LORA_MAPPING = {
    "default": HQQLinearLoRA,
    "lora_with_fakequant": HQQLinearLoRAWithFakeQuant,
    "grouped_proj": HQQLinearGroupedProj,
}


def is_hqq_lora_layer(layer):
    return type(layer) in _HQQ_LORA_CLASSES


##################################################################################################################
def autoname_modules(model):
    for name, module in model.named_modules():
        module.name = name


# Patching functions
def patch_linear_add_peft(layer: nn.Module, patch_params: Union[dict,None]) -> nn.Module:
    _peft_config = patch_params
    if _peft_config:
        lora_type = (
            _peft_config["lora_type"] if ("lora_type" in _peft_config) else "default"
        )
        new_layer = _HQQ_LORA_MAPPING[lora_type](layer, _peft_config)
    else:
        new_layer = layer
    return new_layer


def patch_linear_merge_peft(layer: nn.Module, patch_params: Union[dict,None]) -> nn.Module:
    _quant_config = patch_params
    if _quant_config:
        new_layer = layer.merge_and_quantize(_quant_config)
        del layer
        cleanup()
    else:
        new_layer = layer
    return new_layer


def patch_linear_cast_peft(layer: nn.Module, patch_params: Union[dict,None]) -> nn.Module:
    if is_hqq_lora_layer(layer):
        layer.cast(patch_params)
    return layer


# Putting it all together
class PeftUtils:
    @classmethod
    def get_base_class(cls, model, base_class):
        # Get base class
        if base_class is None:
            base_class = getattr(model, "base_class", None)

        assert (
            base_class is not None
        ), "You need to provide the base HQQ class (LlamaHQQ, MixtralHQQ, etc.) as model.base_class or as an argument base_class=LlamaHQQ. Use AutoHQQHFMode if the model architecture is not supported."

        return base_class

    @classmethod
    def add_lora(
        cls, model, peft_config, base_class: bool = None, verbose: bool = True
    ) -> None:
        # Base classs
        base_class = cls.get_base_class(model, base_class)

        base_class.setup_model(model)

        # Freeze
        for param in model.parameters():
            param.requires_grad = False

        # Patch
        base_class.patch_linearlayers(
            model, patch_linear_add_peft, peft_config, verbose=verbose
        )

        # Rename modules
        autoname_modules(model)

        # add config to model
        model.peft_config = peft_config

    @classmethod
    def merge_lora(
        cls, model, merge_lora_params, base_class: bool = None, verbose: bool = True
    ) -> None:
        # Base classs
        base_class = cls.get_base_class(model, base_class)

        # Patch
        base_class.patch_linearlayers(
            model, patch_linear_merge_peft, merge_lora_params, verbose=verbose
        )

    @classmethod
    def cast_lora_weights(
        cls, model, dtype: torch.dtype, base_class: bool = None, verbose: bool = True
    ) -> None:
        # Base classs
        base_class = cls.get_base_class(model, base_class)

        base_class.setup_model(model)

        # Linear tags
        linear_tags = model.linear_tags

        # Patch
        base_class.patch_linearlayers(
            model,
            patch_linear_cast_peft,
            dict([(linear_tag, dtype) for linear_tag in linear_tags]),
            verbose=verbose,
        )

    @classmethod
    def save_lora_weights(
        cls, model, filename: str, base_class: bool = None, verbose: bool = True
    ) -> None:
        # Base classs
        base_class = cls.get_base_class(model, base_class)

        base_class.setup_model(model)

        lora_global_params = {}

        def _patch_linear_save_weights(layer, patch_params, return_layer=True):
            if is_hqq_lora_layer(layer):
                lora_global_params[layer.name] = layer.state_dict()
            if return_layer:
                return layer

        # Linear tags
        linear_tags = model.linear_tags

        # Patch
        base_class.patch_linearlayers(
            model,
            _patch_linear_save_weights,
            dict([(linear_tag, None) for linear_tag in linear_tags]),
            verbose=verbose,
        )

        # save
        torch.save(
            {"peft_config": model.peft_config, "parameters": lora_global_params},
            filename,
        )

    @classmethod
    def load_lora_weights(
        cls, model, filename: str, base_class: bool = None, verbose: bool = True
    ) -> None:
        # Base classs
        base_class = cls.get_base_class(model, base_class)

        base_class.setup_model(model)

        lora_data = torch.load(filename, map_location="cpu")

        # V0.2 format: automatically creates lora modules if the model doesn't contain it
        if ("peft_config" in lora_data) and ("parameters" in lora_data):
            peft_config = lora_data["peft_config"]
            lora_global_params = lora_data["parameters"]
            if not hasattr(model, "peft_config"):
                cls.add_lora(model=model, peft_config=peft_config)

        # v0.1 format
        else:
            if not hasattr(model, "peft_config"):
                raise Exception(
                    "Using older version of lora weights. LoRa modules should be manually added in this case."
                )
            lora_global_params = lora_data

        def _patch_linear_load_weights(layer, patch_params, return_layer=True):
            if is_hqq_lora_layer(layer):
                layer.load_state_dict(lora_global_params[layer.name])
            if return_layer:
                return layer

        # Linear tags
        linear_tags = model.linear_tags

        # Patch
        base_class.patch_linearlayers(
            model,
            _patch_linear_load_weights,
            dict([(linear_tag, None) for linear_tag in linear_tags]),
            verbose=verbose,
        )



# Written by Dr. Hicham Badri @Mobius Labs GmbH - 2023
#####################################################
from ..base import BasePatch
from .base import BaseHQQHFModel
from tqdm import tqdm


# Patch  functions
class MixtralPatch(BasePatch):
    # These tags are used to specify the parameters of each layer type. For example, if you want to give different quantization parameters to different layers
    @classmethod
    def get_linear_tags(cls):
        return [
            "self_attn.q_proj",
            "self_attn.k_proj",
            "self_attn.v_proj",
            "self_attn.o_proj",
            "block_sparse_moe.experts.w1",
            "block_sparse_moe.experts.w2",
            "block_sparse_moe.experts.w3",
        ]

    @classmethod
    def patch_nonlinearlayers(cls, model, patch_fct, verbose=True):
        base_model = model.model
        model.lm_head = patch_fct(model.lm_head)  ###
        base_model.embed_tokens = patch_fct(base_model.embed_tokens)
        base_model.norm = patch_fct(base_model.norm)

        layers = base_model.layers
        for i in tqdm(range(len(base_model.layers)), disable=not verbose):
            layers[i].self_attn.rotary_emb = patch_fct(layers[i].self_attn.rotary_emb)
            layers[i].input_layernorm = patch_fct(layers[i].input_layernorm)
            layers[i].post_attention_layernorm = patch_fct(
                layers[i].post_attention_layernorm
            )

            layers[i].block_sparse_moe.gate = patch_fct(
                layers[i].block_sparse_moe.gate
            )  # Keep MOE gate as fp16 because it's small

            n_experts = len(layers[i].block_sparse_moe.experts)
            for k in range(n_experts):
                layers[i].block_sparse_moe.experts[k].act_fn = patch_fct(
                    layers[i].block_sparse_moe.experts[k].act_fn
                )

    @classmethod
    def patch_linearlayers(cls, model, patch_fct, patch_params, verbose=True):
        base_model = model.model
        layers = base_model.layers
        for i in tqdm(range(len(layers)), disable=not verbose):
            layers[i].self_attn.q_proj = patch_fct(
                layers[i].self_attn.q_proj, patch_params["self_attn.q_proj"]
            )
            layers[i].self_attn.k_proj = patch_fct(
                layers[i].self_attn.k_proj, patch_params["self_attn.k_proj"]
            )
            layers[i].self_attn.v_proj = patch_fct(
                layers[i].self_attn.v_proj, patch_params["self_attn.v_proj"]
            )
            layers[i].self_attn.o_proj = patch_fct(
                layers[i].self_attn.o_proj, patch_params["self_attn.o_proj"]
            )

            n_experts = len(layers[i].block_sparse_moe.experts)
            for k in range(n_experts):
                layers[i].block_sparse_moe.experts[k].w1 = patch_fct(
                    layers[i].block_sparse_moe.experts[k].w1,
                    patch_params["block_sparse_moe.experts.w1"],
                )
                layers[i].block_sparse_moe.experts[k].w2 = patch_fct(
                    layers[i].block_sparse_moe.experts[k].w2,
                    patch_params["block_sparse_moe.experts.w2"],
                )
                layers[i].block_sparse_moe.experts[k].w3 = patch_fct(
                    layers[i].block_sparse_moe.experts[k].w3,
                    patch_params["block_sparse_moe.experts.w3"],
                )


class MixtralHQQ(MixtralPatch, BaseHQQHFModel):
    pass



import unittest
import torch
from hqq.core.bitpack import BitPack
try:
    import hqq_aten
except:
    print('hqq_aten package not available')
    hqq_aten = None

# python -m unittest test_bitpack.py
class TestBitPack(unittest.TestCase):

    def setUp(self) -> None:
        # set seed
        torch.manual_seed(42)
        self.device  = 'cuda'
        self.repeats = 10
        self.shapes = [[32, 32], [128, 256], [256, 256], [512, 512], [1024, 1024], [2048, 2048], [4096, 4096], [8192, 8192], [8192, 4096], [8192, 128], [32, 4096]]
        return super().setUp()
    
    def tearDown(self) -> None:
        return super().tearDown()

    def test_nbit_torch(self, nbits=None, pack_fct=None, unpack_fct=None, device='cuda'):
        if(nbits is None): 
            self.assertTrue(True)
            return

        for compute_dtype in [torch.float16, torch.bfloat16, torch.float32]:
            for shape in self.shapes:
                for _ in range(self.repeats):
                    W   = torch.randint(0, 2**nbits, shape, device=device).contiguous()
                    W_r = unpack_fct(pack_fct(W), dtype=compute_dtype)                    
                    self.assertTrue(torch.equal(W, W_r[:len(W)]))


    def test_nbit_aten(self, nbits=None, pack_fct=None, unpack_fct=None, device='cuda'):
        if(nbits is None): 
            self.assertTrue(True)
            return

        for compute_dtype in [torch.float16, torch.bfloat16, torch.float32]:
            for shape in self.shapes:
                for _ in range(self.repeats):
                    W   = torch.randint(0, 2**nbits, shape, device=device).contiguous()
                    W_r = unpack_fct(pack_fct(W)).to(compute_dtype)                     
                    self.assertTrue(torch.equal(W, W_r[:len(W)]))

    #8-bit
    def test_8bit_u8_torch(self):
        return self.test_nbit_torch(nbits=8, pack_fct=BitPack.pack_8bit_u8, unpack_fct=BitPack.unpack_8bit_u8, device=self.device)

    #4-bit
    def test_4bit_u8_torch(self):
        return self.test_nbit_torch(nbits=4, pack_fct=BitPack.pack_4bit_u8, unpack_fct=BitPack.unpack_4bit_u8, device=self.device)

    @unittest.skipIf(hqq_aten is None, "hqq_aten not available")
    def test_4bit_u8_aten(self):
        return self.test_nbit_aten(nbits=4, pack_fct=BitPack.pack_4bit_u8, unpack_fct=hqq_aten.unpack_4bit_u8, device=self.device)
    
    #2-bit
    def test_2bit_u8_torch(self):
        return self.test_nbit_torch(nbits=2, pack_fct=BitPack.pack_2bit_u8, unpack_fct=BitPack.unpack_2bit_u8, device=self.device)

    @unittest.skipIf(hqq_aten is None, "hqq_aten not available")
    def test_2bit_u8_aten(self):
        return self.test_nbit_aten(nbits=2, pack_fct=BitPack.pack_2bit_u8, unpack_fct=hqq_aten.unpack_2bit_u8, device=self.device)

    #1-bit
    def test_1bit_u8_torch(self):
        return self.test_nbit_torch(nbits=1, pack_fct=BitPack.pack_1bit_u8, unpack_fct=BitPack.unpack_1bit_u8, device=self.device)

    @unittest.skipIf(hqq_aten is None, "hqq_aten not available")
    def test_1bit_u8_aten(self):
        return self.test_nbit_aten(nbits=1, pack_fct=BitPack.pack_1bit_u8, unpack_fct=hqq_aten.unpack_1bit_u8, device=self.device)

    #3-bit
    def test_3bit_32_torch(self):
        return self.test_nbit_torch(nbits=3, pack_fct=BitPack.pack_3bit_32, unpack_fct=BitPack.unpack_3bit_32, device=self.device)

    @unittest.skipIf(hqq_aten is None, "hqq_aten not available")
    def test_3bit_32_aten(self):
        return self.test_nbit_aten(nbits=3, pack_fct=BitPack.pack_3bit_32, unpack_fct=hqq_aten.unpack_3bit_32, device=self.device)
                                    
if __name__ == '__main__':
    unittest.main()



# Written by Dr. Hicham Badri @Mobius Labs GmbH - 2023
#####################################################
import transformers
from accelerate import init_empty_weights
from ..base import BaseHQQModel, BasePatch


class BaseHQQHFModel(BaseHQQModel):
    # Save model architecture
    @classmethod
    def cache_model(cls, model, save_dir):
        model.config.save_pretrained(save_dir)

    # Create empty model from config
    @classmethod
    def create_model(cls, save_dir, kwargs):
        config_kwargs = {}
        for key in ["attn_implementation"]:
            if key in kwargs:
                config_kwargs[key] = kwargs[key]

        config = transformers.AutoConfig.from_pretrained(
            cls.get_config_file(save_dir), **config_kwargs
        )

        auto_class = transformers.AutoModel

        # Todo: add support for other auto models
        archs = config.architectures
        if len(archs) == 1 and ("CausalLM" in archs[0]):
            auto_class = transformers.AutoModelForCausalLM

        with init_empty_weights():
            model = auto_class.from_config(config)

        return model


# Auto class used for HF models if no architecture was manually setup
class AutoHQQHFModel(BaseHQQHFModel, BasePatch):
    pass



# Written by Dr. Hicham Badri @Mobius Labs GmbH - 2024
#####################################################
import torch

try:
    import marlin
except Exception:
    marlin = None
from ..core.quantize import Quantizer


class MarlinLinear(torch.nn.Module):
    def __init__(
        self, W: torch.Tensor, scales: torch.Tensor, u=None, bias=None, groupsize=-1
    ):
        super().__init__()

        m, n = W.shape
        device = W.device
        _linear = torch.nn.Linear(m, n)
        _linear.weight.data = W.half().t()

        effective_groupsize = m if (groupsize == -1) else groupsize

        _layer = marlin.Layer(m, n, groupsize=groupsize)
        _layer.k = m
        _layer.n = n
        _layer.groupsize = effective_groupsize
        _layer.B = torch.empty((m // 16, n * 16 // 8), dtype=torch.int, device=device)
        _layer.s = torch.empty(
            (m // effective_groupsize, n), dtype=torch.half, device=device
        )
        _layer.pack(_linear, scales.t())

        self.bias = bias.half() if (bias is not None) else None
        self.Wq_packed = _layer.B.clone()
        self.scales = _layer.s.clone()
        self.workspace_fp = torch.zeros(n // 128 * 16, device=device)
        self.in_features = m
        self.out_features = n
        self.group_size = effective_groupsize
        self.axis = 1
        self.device = device
        self.compute_dtype = torch.float16
        self.u = torch.nn.Parameter(u, requires_grad=False) if (u is not None) else None

        del _linear, _layer
        torch.cuda.empty_cache()

    @torch.no_grad()
    def matmul(self, x):
        out = torch.empty(
            x.shape[:-1] + (self.scales.shape[1],), dtype=x.dtype, device=x.device
        )
        marlin.mul(
            x.view((-1, x.shape[-1])),
            self.Wq_packed,
            out.view((-1, out.shape[-1])),
            self.scales,
            self.workspace_fp,
        )
        return out

    @torch.jit.ignore
    def forward(self, x):
        out = self.matmul(x)

        if self.u is not None:
            out += torch.matmul(x.sum(axis=-1, keepdim=True), self.u)

        if self.bias is not None:
            out += self.bias

        return out


# ONLY WORKS WITH AXIS=1, group_size= - 1
def patch_hqq_to_marlin(layer, patch_params):
    if marlin is None:
        return layer

    z_shift = 8.0
    hqq_layer = layer.linear_layer if hasattr(layer, "linear_layer") else layer

    # Check config suppport
    if (
        (hqq_layer.meta["axis"] == 0)
        or (hqq_layer.meta["group_size"] is not None)
        or (hqq_layer.meta["nbits"] != 4)
    ):
        print("Skipping marlin conversion for", hqq_layer.name)
        return layer

    W_r = Quantizer.unpack[hqq_layer.meta["packing"]](
        hqq_layer.W_q, dtype=hqq_layer.compute_dtype
    ).t()
    z = hqq_layer.meta["zero"]
    s = hqq_layer.meta["scale"].t()
    W_r = (W_r - z_shift) * s

    if type(z) in [torch.Tensor, torch.nn.Parameter]:
        z = z.t()
        u = (s * (-z + z_shift)).view([1, -1])
    else:
        u = None

    marlin_layer = MarlinLinear(W_r, s, u=u, bias=hqq_layer.bias)

    if hasattr(layer, "linear_layer"):
        del layer.linear_layer
        layer.linear_layer = marlin_layer
    else:
        del hqq_layer
        layer = marlin_layer

    torch.cuda.empty_cache()

    return layer



from ..base import BaseHQQModel
import os
import json
import timm


class BaseHQQTimmModel(BaseHQQModel):
    # Save model architecture
    @classmethod
    def cache_model(cls, model, save_dir, kwargs):
        try:
            os.makedirs(save_dir, exist_ok=True)
        except Exception as error:
            print(error)

        with open(cls.get_config_file(save_dir), "w") as file:
            json.dump(model.default_cfg, file)

    # Create empty model
    @classmethod
    def create_model(cls, save_dir):
        with open(cls.get_config_file(save_dir), "r") as file:
            config = json.load(file)
        model = timm.create_model(
            config["architecture"] + "." + config["tag"], pretrained=False
        )
        return model



# Written by Dr. Hicham Badri @Mobius Labs GmbH - 2023
#####################################################
from abc import abstractmethod
from ..models.base import BaseHQQModel
import torch
from torch import float16
from typing import Union

# Wrapper that makes it easier to add quantization support to different engines (HF, VLLM, etc.)


class HQQWrapper:
    @abstractmethod
    def _get_arch_key_from_save_dir(cls, save_dir: str):
        pass

    @classmethod
    def _get_hqq_class(cls, arg):
        arch = arg if (isinstance(arg, str)) else arg.arch_key
        return cls._HQQ_REGISTRY[arch]

    @classmethod
    def _validate_params(cls, params: dict):
        pass

    @classmethod
    def _is_quantizable(cls, model):
        return hasattr(model, "hqq_quantized")

    @classmethod
    def _make_quantizable(cls, model, quantized: bool):
        model.hqq_quantized = quantized
        model.base_class = cls._get_hqq_class(model)

    @classmethod
    def _check_arch_support(cls, arg):
        arch = arg if (isinstance(arg, str)) else arg.arch_key
        assert arch in cls._HQQ_REGISTRY, (
            "Model architecture " + arch + " not supported yet."
        )

    @classmethod
    def _check_if_already_quantized(cls, model):
        assert not model.hqq_quantized, "Model already quantized"

    @classmethod
    def _check_if_not_quantized(cls, model):
        assert model.hqq_quantized, "Model not quantized."

    @classmethod
    def _set_quantized(cls, model, quantized: bool):
        model.hqq_quantized = quantized

    #####################################################
    @classmethod
    def quantize_model_(
        cls, model, quant_config, compute_dtype: torch.dtype = float16, device="cuda"
    ):
        if cls._is_quantizable(model) is False:
            cls._make_quantizable(model, quantized=False)
        cls._check_arch_support(model)
        cls._check_if_already_quantized(model)
        cls._get_hqq_class(model).quantize_model(
            model, quant_config=quant_config, compute_dtype=compute_dtype, device=device
        )
        cls._set_quantized(model, True)

    @classmethod
    def save_quantized_(cls, model, save_dir: str):
        cls._check_if_not_quantized(model)
        cls._get_hqq_class(model).save_quantized(model, save_dir=save_dir)

    @classmethod
    def from_quantized(
        cls,
        save_dir_or_hub,
        compute_dtype: torch.dtype = float16,
        device="cuda",
        cache_dir: Union[str,None] = "",
        adapter: str = None,
    ):
        # Both local and hub-support
        save_dir = BaseHQQModel.try_snapshot_download(save_dir_or_hub, cache_dir=cache_dir)
        arch_key = cls._get_arch_key_from_save_dir(save_dir)
        cls._check_arch_support(arch_key)

        model = cls._get_hqq_class(arch_key).from_quantized(
            save_dir,
            compute_dtype=compute_dtype,
            device=device,
            cache_dir=cache_dir,
            adapter=adapter,
        )

        cls._make_quantizable(model, quantized=True)
        return model

    @classmethod
    def get_linear_tags(cls, model):
        return cls._get_hqq_class(model).get_linear_tags()



# Written by Dr. Hicham Badri @Mobius Labs GmbH - 2023
#####################################################

import transformers
import torch
from .base import HQQWrapper

from ..models.hf.llama import LlamaHQQ
from hqq.models.hf.mixtral import MixtralHQQ
from hqq.models.hf.phi import PhiHQQ
from hqq.models.hf.mistral import MistralHQQ

_HQQ_REGISTRY = {}
_HQQ_REGISTRY["LlamaForCausalLM"] = LlamaHQQ
_HQQ_REGISTRY["MixtralForCausalLM"] = MixtralHQQ
_HQQ_REGISTRY["PhiForCausalLM"] = PhiHQQ
_HQQ_REGISTRY["MistralForCausalLM"] = MistralHQQ

# Alias
AutoTokenizer = transformers.AutoTokenizer

# Used to call super() on classmethods
_Parent = transformers.AutoModelForCausalLM


class HQQModelForCausalLM(_Parent, HQQWrapper):
    _HQQ_REGISTRY = _HQQ_REGISTRY

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def _make_quantizable(cls, model, quantized: bool) -> None:
        model.hqq_quantized = quantized
        model.arch_key = model.config.architectures[0]
        model.quantize_model = (
            lambda quant_config,
            compute_dtype=torch.float16,
            device="cuda": cls.quantize_model_(
                model=model,
                quant_config=quant_config,
                compute_dtype=compute_dtype,
                device=device,
            )
        )
        model.save_quantized = lambda save_dir: cls.save_quantized_(
            model=model, save_dir=save_dir
        )
        model.cuda = lambda *args, **kwargs: model if (quantized) else model.cuda
        model.to = lambda *args, **kwargs: model if (quantized) else model.to
        model.float = lambda *args, **kwargs: model if (quantized) else model.float
        model.half = lambda *args, **kwargs: model if (quantized) else model.half
        model.base_class = cls._get_hqq_class(model)

    # Force loading the model on CPU and unquantized
    @classmethod
    def _validate_params(cls, params: dict) -> None:
        for p in ["load_in_4bit", "load_in_8bit"]:  # ignore these
            if p in params:
                params[p] = False
        params["device_map"] = None

    @classmethod
    def from_pretrained(cls, *args, **kwargs):
        cls._validate_params(kwargs)
        model = super(_Parent, cls).from_pretrained(*args, **kwargs)
        cls._make_quantizable(model, quantized=False)
        return model

    @classmethod
    def _get_arch_key_from_save_dir(cls, save_dir: str):
        config = transformers.AutoConfig.from_pretrained(save_dir)
        return config.architectures[0]



from setuptools import setup
from torch.utils import cpp_extension

setup(
    name="hqq_aten",
    ext_modules=[cpp_extension.CppExtension("hqq_aten", ["hqq_aten.cpp"])],
    extra_compile_args=["-O3"],
    cmdclass={"build_ext": cpp_extension.BuildExtension},
)

# python3 setup.py install



from ..base import BasePatch
from .base import BaseHQQTimmModel
from tqdm import tqdm


# Patch ViT functions
class VitCLIPPatch(BasePatch):
    # These tags are used to specify the parameters of each layer type. For example, if you want to give different quantization parameters to different layers
    @classmethod
    def get_linear_tags(cls):
        return ["attn.qkv", "attn.proj", "mlp.fc1", "mlp.fc2"]

    @classmethod
    def freeze_model(cls, model):
        for param in model.parameters():
            param.requires_grad = False

    @classmethod
    def patch_nonlinearlayers(cls, model, patch_fct, verbose=True):
        model.patch_embed.proj = patch_fct(model.patch_embed.proj)
        model.patch_embed.norm = patch_fct(model.patch_embed.norm)
        model.norm_pre = patch_fct(model.norm_pre)
        model.norm = patch_fct(model.norm)
        model.head = patch_fct(model.head)
        model.cls_token.data = patch_fct(model.cls_token.data)
        model.pos_embed.data = patch_fct(model.pos_embed.data)

        for i in tqdm(range(len(model.blocks)), disable=not verbose):
            model.blocks[i].norm1 = patch_fct(model.blocks[i].norm1)
            model.blocks[i].norm2 = patch_fct(model.blocks[i].norm2)

    @classmethod
    def patch_linearlayers(cls, model, patch_fct, patch_params, verbose=True):
        for i in tqdm(range(len(model.blocks))):
            model.blocks[i].attn.qkv = patch_fct(
                model.blocks[i].attn.qkv, patch_params["attn.qkv"]
            )
            model.blocks[i].attn.proj = patch_fct(
                model.blocks[i].attn.proj, patch_params["attn.proj"]
            )
            model.blocks[i].mlp.fc1 = patch_fct(
                model.blocks[i].mlp.fc1, patch_params["mlp.fc1"]
            )
            model.blocks[i].mlp.fc2 = patch_fct(
                model.blocks[i].mlp.fc2, patch_params["mlp.fc2"]
            )


class ViTCLIPHQQ(VitCLIPPatch, BaseHQQTimmModel):
    # layers to ignore when saving the weights
    @classmethod
    def get_ignore_layers(cls, model):
        return ["", "model", "model.blocks"] + [
            "model.blocks." + str(i) for i in range(len(model.blocks))
        ]

    # since cls_token and pos_embed are trainable parameters but are not part of any module, we need to add them manually
    # for saving
    @classmethod
    def serialize_weights(cls, model, verbose):
        weights = super().serialize_weights(model, verbose)
        weights["cls_token"] = model.cls_token.data
        weights["pos_embed"] = model.pos_embed.data
        return weights

    # and loading
    @classmethod
    def post_module_load(cls, model, weights):
        super().post_module_load(model, weights)
        model.cls_token.data = weights["cls_token"]
        model.pos_embed.data = weights["pos_embed"]



# Written by Dr. Hicham Badri @Mobius Labs GmbH - 2023
#####################################################
import os
import torch
from torch import nn
from torch import float16
from os.path import join as pjoin
from typing import Callable
from tqdm import tqdm
from abc import abstractmethod
from functools import partial
from typing import Union

from huggingface_hub import snapshot_download
from ..core.utils import cleanup
from ..core.quantize import HQQLinear
from ..core.peft import PeftUtils, _HQQ_LORA_CLASSES


# Defined what is qualified as "linear layer"
_QUANT_LAYERS = [nn.Linear, HQQLinear] + _HQQ_LORA_CLASSES
_IGNORE_LINEAR = ["lm_head"]


# Finds the parent of a node module named "name"
def find_parent(model, name: str) -> nn.Module:
    module_tree = name.split(".")[:-1]
    parent = model
    for m in module_tree:
        parent = parent._modules[m]
    return parent


# checks if a module is a leaf: doesn't have another module inside
def is_leaf_module(module) -> bool:
    return len(module._modules) == 0


# Get the linear_tag from a modul name. For example: model.layers.31.self_attn.k_proj -> self_attn.k_proj
def name_to_linear_tag(name: str) -> str:
    return ".".join(
        [
            n
            for n in name.split(".")
            if ((n not in ["model", "layers"]) and (not n.isnumeric()))
        ]
    )


# returns all children nodes from model
def get_all_children_from_model(model, ignore: list = []) -> list:
    tags = []
    for name, module in model.named_modules():
        if is_leaf_module(module) and (name.split(".")[-1] not in ignore):
            tags.append(name)
    return tags


# Get all linear tags available
def get_linear_tags_from_model(model, ignore: list) -> list:
    linear_tags = set()
    for name, module in model.named_modules():
        if (type(module) in _QUANT_LAYERS) and (name.split(".")[-1] not in ignore):
            linear_tags.add(name_to_linear_tag(name))
    return list(linear_tags)


def forward_device_hooked(self, *args, **kwargs):
    args = list(args)

    # eddit this to make torch.compile compatible
    for i in range(len(args)):
        if isinstance(
            args[i], (torch.Tensor, torch.nn.Parameter)
        ):  # if hasattr(args[i], "to"):
            args[i] = args[i].to(self.device)

    for i in kwargs:
        if isinstance(
            kwargs[i], (torch.Tensor, torch.nn.Parameter)
        ):  # if hasattr(kwargs[i], "to"):
            kwargs[i] = kwargs[i].to(self.device)

    # return self.__class__.forward(self, *args, **kwargs)
    return self.forward_orig(*args, **kwargs)


# Base patching class. Patching defines how nn.Linear and other layers are replaced via a patching function.
class BasePatch:
    # Override these OR override the main patch_model() function
    ############################################
    # This method iterates through layers of the model that are NOT nn.Linear and processes them via new_nodule = patch_fct(module, params)
    @classmethod
    def patch_nonlinearlayers(
        cls, model, patch_fct: Callable, verbose: bool = True
    ) -> None:
        ignore_tags = cls.get_ignore_layers(model)

        tmp_mapping = {}
        for name, module in model.named_modules():
            if (type(module) not in _QUANT_LAYERS) and (name not in ignore_tags):
                tmp_mapping[name] = module

        for name in tqdm(tmp_mapping, disable=not verbose):
            setattr(
                find_parent(model, name),
                name.split(".")[-1],
                patch_fct(tmp_mapping[name]),
            )

        cleanup()

    # This method iterates through layers of the model that are nn.Linear and processes them via new_nodule = patch_fct(module, params)
    @classmethod
    def patch_linearlayers(
        cls,
        model,
        patch_fct: Callable,
        patch_params: Union[dict, None],
        verbose: bool = True,
    ) -> None:
        ignore_tags = cls.get_ignore_layers(model)

        tmp_mapping = {}
        for name, module in model.named_modules():
            if (type(module) in _QUANT_LAYERS) and (name not in ignore_tags):
                tmp_mapping[name] = module

        for name in tqdm(tmp_mapping, disable=not verbose):
            linear_tag = name_to_linear_tag(name)
            patch_param = (
                patch_params[linear_tag] if (linear_tag in patch_params) else None
            )
            setattr(
                find_parent(model, name),
                name.split(".")[-1],
                patch_fct(tmp_mapping[name], patch_param),
            )

        cleanup()

    ############################################
    # These tags are used to specfiy parameters of the patching in patch_linearlayers()
    @classmethod
    def set_auto_linear_tags(cls, model, ignore: list = _IGNORE_LINEAR) -> None:
        if hasattr(model, "linear_tags") is False:
            linear_tags = cls.get_linear_tags()
            model.linear_tags = (
                linear_tags
                if len(linear_tags) > 0
                else get_linear_tags_from_model(model, ignore=ignore)
            )
            model.base_class = cls

    # Returns the current linear tags
    @classmethod
    def get_linear_tags(cls) -> list:
        return []

    @classmethod
    def get_ignore_layers(cls, model) -> list:
        layers = {""}
        for name, module in model.named_modules():
            if not is_leaf_module(module):
                layers.add(name)
        return list(layers)

    # Autmatically name modules. This is very important to save/load the weights
    @classmethod
    def autoname_modules(cls, model) -> None:
        for name, module in model.named_modules():
            module.name = name

    # Freeze all layers
    @classmethod
    def freeze_model(cls, model) -> None:
        for param in model.parameters():
            param.requires_grad = False
        try:
            for param in model.model.parameters():
                param.requires_grad = False
        except Exception:
            pass

    # Main patching function
    @classmethod
    def patch_model(
        cls,
        model,
        patch_nonlinear_fct: Callable,
        patch_linear_fct: Callable,
        patch_params: dict,
        verbose: bool = True,
    ) -> None:
        model.eval()
        cls.freeze_model(model)
        cls.autoname_modules(model)
        cls.patch_nonlinearlayers(model, patch_nonlinear_fct, verbose=verbose)
        cls.patch_linearlayers(model, patch_linear_fct, patch_params, verbose=verbose)
        cleanup()


class BaseHQQModel:
    # Override these
    ############################################
    # This method creates and empty model based on the specfied architecture
    @abstractmethod
    def create_model(cls, save_dir, kwargs):
        pass

    # This method saves the model architecture only without inculding the weights (for example to a config.json)
    @abstractmethod
    def cache_model(cls, model, save_dir: str):
        pass

    ############################################

    @classmethod
    def get_config_file(cls, save_dir: str) -> str:
        return pjoin(save_dir, "config.json")

    @classmethod
    def get_weight_file(cls, save_dir: str) -> str:
        return pjoin(save_dir, "qmodel.pt")

    # Save weights to disk
    @classmethod
    def save_weights(cls, weights: dict, save_dir: str) -> None:
        torch.save(weights, cls.get_weight_file(save_dir))

    # Load weights from disk
    @classmethod
    def load_weights(cls, save_dir: str, map_location=None):
        return torch.load(cls.get_weight_file(save_dir), map_location=map_location)

    # Set-up model with the necessary data
    @classmethod
    def setup_model(cls, model):
        cls.autoname_modules(model)
        cls.set_auto_linear_tags(model)

    # Main function to quantize a model. Basically goes through the linear layers specfied in the patching function and replaces them with HQQLinear
    @classmethod
    def quantize_model(
        cls,
        model,
        quant_config: dict,
        compute_dtype: torch.dtype = float16,
        device: Union[str, list, dict] = "cuda",
    ):
        # Check if the model was already quantized
        if getattr(model, "hqq_quantized", False):
            print("Model was already quantized")
            return

        # Set linear tags automatically
        cls.setup_model(model)

        # Use the same quantization config for all linear layers. Use None to skip quantizing a specfic layer.
        if True in [(key in model.linear_tags) for key in quant_config.keys()]:
            # If the user doesn't specify a key from get_linear_tags, the layer is not quantized via (key, None)
            patch_params = {key: None for key in model.linear_tags}
            patch_params.update(quant_config)
        else:
            # Same quant_config for all layers
            patch_params = {k: quant_config for k in model.linear_tags}

        # Get list of all nodes in order
        all_nodes = get_all_children_from_model(model, [])  # ordered nodes
        try:
            # Extract block names: This is following Hugging Face models.
            num_blocks = (
                len(model.model.layers)
                if hasattr(model, "model")
                else len(model.layers)
            )
            all_blocks = ["model.layers." + str(i) for i in range(num_blocks)]
        except Exception:
            all_blocks = None
            print(
                "Default model structure not supported. Make sure you feed device as dictionary as {name_block: device}"
            )

        if isinstance(
            device, dict
        ):  # input as {module block name (str): device (str or torch.device)}
            device_map = device
            num_devices = len(set([device_map[k] for k in device_map]))
            all_blocks = list(device_map.keys())

        node_to_block = {}
        for node in all_nodes:
            res = [block for block in all_blocks if (block in node)]
            node_to_block[node] = res[-1] if (len(res) > 0) else node

        # Set device-map
        if isinstance(device, str):  # single device as str
            device_map = {k: device for k in all_blocks + all_nodes}
            num_devices = 1

        if isinstance(device, list):  # list of devices
            num_devices = len(device)
            device_map = {}
            for node in all_nodes:
                if ".layers" in node:
                    break
                device_map[node] = device[0]

            for node in all_nodes[::-1]:
                if ".layers" in node:
                    break
                device_map[node] = device[-1]

            step, k = len(all_blocks) // num_devices, 0
            for i in range(0, len(all_blocks), step):
                for j in range(i, i + step):
                    device_map[all_blocks[min(j, len(all_blocks) - 1)]] = device[
                        min(k, num_devices - 1)
                    ]
                k += 1

        # Map nodes to block devices
        for node in all_nodes:
            device_map[node] = device_map[node_to_block[node]]

        # We replace the nn.Linear layers with HQQLinear
        def _patch_linear(linear_layer, quant_config):
            if type(linear_layer) is HQQLinear:
                return linear_layer

            current_device = device_map[linear_layer.name]

            if quant_config is not None:
                out_module = HQQLinear(
                    linear_layer,
                    quant_config,
                    compute_dtype=compute_dtype,
                    device=current_device,
                )
            else:
                out_module = linear_layer.to(device=current_device, dtype=compute_dtype)

            out_module.device = current_device
            return out_module

        def _patch_other(layer):
            current_device = device_map[layer.name]
            layer.device = current_device
            return layer.to(device=current_device, dtype=compute_dtype)

        cls.patch_model(model, _patch_other, _patch_linear, patch_params)

        # Insert device switcher
        if num_devices > 1:
            core_model = model if hasattr(model, "layers") else model.model

            # Make sure the input (first node) has the input in the right device during generation
            input_node_child_name = all_nodes[0].split(".")[-1]
            input_node = getattr(core_model, input_node_child_name)
            input_node.device = device_map[all_nodes[0]]
            input_node.forward_orig = input_node.forward
            input_node.forward = partial(forward_device_hooked, input_node)
            setattr(core_model, input_node_child_name, input_node)

            # Make sure all inputs to the blocks are in the right device
            for i in range(len(core_model.layers)):
                core_model.layers[i].device = device_map[core_model.layers[i].name]
                core_model.layers[i].forward_orig = core_model.layers[i].forward
                core_model.layers[i].forward = partial(
                    forward_device_hooked, core_model.layers[i]
                )

        # Set base class
        model.base_class = cls

        model.hqq_quantized = True

        return model

    # Prepares model weights by iterating through modules. It might some parameters that are NOT modules like model.param1
    @classmethod
    def serialize_weights(cls, model, verbose: bool = False) -> dict:
        weights = {}
        ignore_keys = cls.get_ignore_layers(model)
        for name, module in model.named_modules():
            if name in ignore_keys:
                continue
            try:
                state_dict = module.state_dict()
                if len(state_dict) > 0:
                    weights[name] = dict(state_dict)
            except Exception:
                if verbose:
                    print("Skipping", name)

        return weights

    # Main function to save a quantized model
    @classmethod
    def save_quantized(cls, model, save_dir: str, verbose: bool = False):
        # Save config
        cls.cache_model(model, save_dir)

        # Serialization
        weights = cls.serialize_weights(model, verbose=verbose)

        # Save
        cls.save_weights(weights, save_dir)

    @classmethod
    def try_snapshot_download(
        cls, save_dir_or_hub: str, cache_dir: Union[str, None] = ""
    ):
        if cache_dir is None:
            save_dir = pjoin("", save_dir_or_hub)
        else:
            save_dir = pjoin(cache_dir, save_dir_or_hub)

        if not os.path.exists(save_dir):
            save_dir = snapshot_download(repo_id=save_dir_or_hub, cache_dir=cache_dir)
            save_dir = pjoin(save_dir)

        # Check
        if not os.path.exists(cls.get_weight_file(save_dir)):
            raise Exception("Weight file missing. Check your cache directory.")
        if not os.path.exists(cls.get_config_file(save_dir)):
            raise Exception("Config file missing. Check your cache directory.")

        return save_dir

    # This method is specfically designed in case we need to load some weights that are not part of any module
    @classmethod
    def post_module_load(cls, model, weights: dict):
        pass

    # Main function to load an HQQ quantized model from either HF hub or locally
    @classmethod
    def from_quantized(
        cls,
        save_dir_or_hub,
        compute_dtype: torch.dtype = float16,
        device="cuda",
        cache_dir: Union[str, None] = "",
        adapter: str = None,
        **kwargs,
    ):
        # Get directory path
        save_dir = cls.try_snapshot_download(save_dir_or_hub, cache_dir)

        # Load model from config
        model = cls.create_model(save_dir, kwargs)

        # Track save directory
        model.save_dir = save_dir

        # Name the layers
        cls.setup_model(model)

        # Load weights
        try:
            weights = cls.load_weights(save_dir)
        except Exception:
            print("Failed to load the weights")
            raise FileNotFoundError

        # load_state_dict() doesn't work with modules initialized with init_empty_weights(), so we need to do this manually
        @torch.no_grad()
        def _load_module(module, params=None):
            if module.name not in weights:
                return module.to(device=device, dtype=compute_dtype, non_blocking=True)

            state_dict = weights[module.name]
            if ("W_q" in state_dict) and ("meta" in state_dict):
                module = HQQLinear(
                    linear_layer=None,
                    quant_config=None,
                    compute_dtype=compute_dtype,
                    device=device,
                )
                module.load_state_dict(state_dict)
            else:
                for key in state_dict:
                    setattr(
                        module,
                        key,
                        nn.Parameter(
                            state_dict[key].to(
                                device=device, dtype=compute_dtype, non_blocking=True
                            ),
                            requires_grad=False,
                        ),
                    )

            return module

        # Load modules
        cls.patch_model(
            model, _load_module, _load_module, {k: None for k in model.linear_tags}
        )

        # Load other weights that are not part of any module
        cls.post_module_load(model, weights)

        model.hqq_quantized = True

        # Set base class
        model.base_class = cls

        # Add adapter
        if adapter is not None:
            try:
                PeftUtils.load_lora_weights(model, filename=pjoin(save_dir, adapter))
                PeftUtils.cast_lora_weights(model, dtype=compute_dtype)
            except Exception as e:
                print("Skipping adapter loading...", str(e))

        return model



# Written by Dr. Hicham Badri @Mobius Labs GmbH - 2024
#####################################################

# Makes HQQ 4-bit (axis=1) compatbile with torch.ops.aten._weight_int4pack_mm

# The code is partially based on: https://github.com/pytorch-labs/gpt-fast/blob/main/quantize.py (BSD-3-Clause license)
# Only the packing logic is copied, the rest is rewritten to support HQQ's logic

# Only works with: bfloat16, axis=1.
# Only tested on Ada gpus.

import torch
import copy
from torch import bfloat16, nn, Tensor
import torch.nn.functional as F
from typing import Union

from ..core.quantize import HQQLinear, Quantizer
from ..core.peft import HQQLinearLoRA
from ..core.utils import cleanup


class HQQLinearTorchWeightOnlynt4(torch.nn.Module):
    def __init__(
        self,
        linear_layer: Union[nn.Module, None],
        quant_config: dict,
        del_orig: bool = True,
        compute_dtype: torch.dtype = bfloat16,
        device: str = "cuda",
        initialize: bool = True,
        inner_k_tiles=8,
        padding=True,
    ):
        super().__init__()

        self.ready = False
        self.in_gpu = False
        self.bias = None
        self.device = device
        self.compute_dtype = compute_dtype
        self.quant_config = (
            copy.deepcopy(quant_config) if (quant_config is not None) else None
        )
        self.del_orig = del_orig

        if (quant_config is None) and (linear_layer is None):
            raise Exception(
                "Invalid parameters: Both quant_config and linear_layer are None."
            )

        if (quant_config is None) and (initialize is True):
            raise Exception(
                "Invalid parameters: setting initialize to True requires a quant_config."
            )

        if self.quant_config is not None:
            weight_quant_params = self.quant_config["weight_quant_params"]
            self.groupsize = weight_quant_params["group_size"]
            self.nbits = weight_quant_params["nbits"]
            self.axis = weight_quant_params["axis"]

        if linear_layer is not None:
            self.groupsize = linear_layer.meta["group_size"]
            self.nbits = linear_layer.meta["nbits"]
            self.axis = linear_layer.meta["axis"]

        self.inner_k_tiles = inner_k_tiles
        self.padding = padding

        assert self.axis==1, "Only axis==1 is supported"
        assert self.nbits in [4], "Unsupported nbits."
        assert (
            self.compute_dtype is bfloat16
        ), "Only bfloat16 compute_dtype is supported."
        assert self.groupsize in [None, 32, 64, 128, 256], "Unsupported groupsize."
        assert self.inner_k_tiles in [2, 4, 8], "Unsupported tile."

        self.linear_layer = linear_layer

        if initialize:
            self.initialize()

    ###################### Initializers ######################
    def initialize_with_hqq_quants(self, W_q, meta, bias=None):
        self.padding = (
            False  # Force padding off, a bit tricky to post-pad with grouping
        )

        self.set_shape(meta["shape"])
        self.process_hqq_quants(W_q, meta)
        self.bias = bias
        self.ready = True
        self.in_gpu = True
        torch.cuda.empty_cache()

        return self

    def initialize(self):
        if self.linear_layer is not None:
            W = self.linear_layer.weight.data
            self.set_shape(W.shape)

            if self.in_features_diff > 0:
                W = F.pad(W, pad=(0, self.in_features_diff), value=0)

            W_q, meta = self.quantize(W, **self.quant_config)
            self.process_hqq_quants(W_q, meta)
            del W_q, meta

            self.bias = (
                None
                if (self.linear_layer.bias is None)
                else self.linear_layer.bias.to(
                    dtype=self.compute_dtype, device=self.device
                )
            )

        if self.del_orig:
            del self.linear_layer

        self.ready = True
        self.in_gpu = True
        torch.cuda.empty_cache()

        return self

    ###################### Quantize/packing ######################

    def quantize(
        self,
        W: Tensor,
        weight_quant_params: dict,
        scale_quant_params=Union[dict,None],
        zero_quant_params=Union[dict,None],
        offload_meta=False,
    ):
        W_q, meta = Quantizer.quantize(
            W,
            **weight_quant_params,
            device=self.device,
            compute_dtype=self.compute_dtype,
            bitpack=False,
        )

        # ToDO: meta quantization

        return W_q, meta

    # TODO: move these to utils
    @torch.no_grad()
    def reshape_meta_axis1(self, meta_tensor, new_group_size, shape):
        meta_tensor = meta_tensor.repeat([1, shape[1]]).reshape(shape)
        meta_tensor = torch.mean(
            meta_tensor.reshape([-1, new_group_size]), axis=1, keepdim=True
        )
        return meta_tensor

    def find_multiple(self, n: int, k: int) -> int:
        if n % k == 0:
            return n
        return n + k - (n % k)

    def set_shape(self, shape):
        self.shape = shape
        self.in_features = shape[1]
        self.out_features = shape[0]

        self.origin_in_features = self.in_features
        if self.padding:
            self.in_features = self.find_multiple(self.in_features, 1024)

        self.in_features_diff = self.in_features - self.origin_in_features

    @torch.no_grad()
    def process_hqq_quants(self, W_q, meta):
        scales = meta["scale"]
        zeros = meta["zero"]
        shape = meta["shape"]

        if meta["packing"] is not None:
            W_q = Quantizer.unpack[meta["packing"]](W_q)

        if self.groupsize is None:
            self.groupsize = 128
            W_q = W_q.reshape([-1, self.groupsize])
            scales = self.reshape_meta_axis1(scales, self.groupsize, shape)
            zeros = self.reshape_meta_axis1(zeros, self.groupsize, shape)

        W_q_torch, scales_torch, zeros_torch = self.hqq_quants_to_torch_quants(
            W_q=W_q, scales=scales, zeros=zeros, shape=shape, nbits=self.nbits
        )
        self.weight_int4pack = torch.ops.aten._convert_weight_to_int4pack(
            W_q_torch, self.inner_k_tiles
        )
        self.scales_and_zeros = self.pack_scales_and_zeros(scales_torch, zeros_torch)

        del W_q_torch, scales_torch, zeros_torch
        torch.cuda.empty_cache()

    @torch.no_grad()
    def hqq_quants_to_torch_quants(
        self, W_q: Tensor, scales: Tensor, zeros: Tensor, shape, nbits=4
    ):
        W_q = W_q.to(dtype=self.compute_dtype, device=self.device)
        scales = scales.to(dtype=self.compute_dtype, device=self.device)
        zeros = zeros.to(dtype=self.compute_dtype, device=self.device)

        max_int = 2**nbits - 1
        min_int = 0
        dump = 2 ** (nbits - 1)

        # HQQ -> torch logic
        new_zeros = (scales * dump) - zeros * scales

        min_val = new_zeros - scales * dump

        # group_quantize_tensor_from_qparams
        W_r = (W_q - zeros) * scales

        W_q = (
            W_r.sub(min_val)
            .div(scales)
            .round()
            .clamp_(min_int, max_int)
            .to(torch.int32)
            .reshape(shape)
            .contiguous()
        )

        # group_dequantize_tensor_from_qparams
        # W_r = W_q*scales + min_val

        scales = scales.contiguous().reshape(shape[0], -1)
        new_zeros = new_zeros.contiguous().reshape(shape[0], -1)

        return W_q, scales, new_zeros

    def pack_scales_and_zeros(self, scales, zeros):
        assert scales.shape == zeros.shape
        assert scales.dtype == bfloat16
        assert zeros.dtype == bfloat16
        return (
            torch.cat(
                [
                    scales.reshape(scales.size(0), scales.size(1), 1),
                    zeros.reshape(zeros.size(0), zeros.size(1), 1),
                ],
                2,
            )
            .transpose(0, 1)
            .contiguous()
        )

    ###################### Forward/matmul ######################

    # @torch.jit.ignore()
    def matmul(self, x: Tensor) -> Tensor:
        origin_x_size = x.size()
        x = x.reshape(-1, origin_x_size[-1])
        c = torch.ops.aten._weight_int4pack_mm(
            x, self.weight_int4pack, self.groupsize, self.scales_and_zeros
        )
        new_shape = origin_x_size[:-1] + (self.out_features,)
        c = c.reshape(new_shape)
        return c

    # TODO without matmul
    def dequantize(self) -> Tensor:
        return self.matmul(
            torch.eye(self.in_features, dtype=self.compute_dtype, device=self.device)
        )[: self.origin_in_features].t()

    # TODO: backward
    def forward(self, x: Tensor) -> Tensor:
        if self.in_features_diff > 0:
            x = F.pad(x, pad=(0, self.in_features_diff))

        out = self.matmul(x)

        if self.bias is not None:
            out += self.bias
        return out


###################### Patching ######################
def patch_linearlayers(model, fct, patch_param=None):
    model.base_class.patch_linearlayers(
        model,
        fct,
        {lin_tag: patch_param for lin_tag in model.base_class.get_linear_tags()},
    )


def patch_hqq_to_aoint4(layer, patch_params):
    hqq_layer = None
    if type(layer) is HQQLinear:
        hqq_layer = layer
    if type(layer) is HQQLinearLoRA:
        hqq_layer = layer.linear_layer

    if hqq_layer is None:
        return layer

    if hqq_layer.meta["nbits"] != 4 or hqq_layer.meta["axis"] != 1:
        print("Skipping aoint4 conversion for ", hqq_layer.name)
        return layer

    quant_config = getattr(hqq_layer, "quant_config", None)

    hqq_aoint4_layer = HQQLinearTorchWeightOnlynt4(
        None,
        quant_config=quant_config,
        compute_dtype=hqq_layer.compute_dtype,
        device=hqq_layer.device,
        del_orig=False,
        initialize=False,
        padding=False,
    )
    hqq_aoint4_layer.initialize_with_hqq_quants(
        hqq_layer.W_q, hqq_layer.meta, hqq_layer.bias
    )

    del hqq_layer
    torch.cuda.empty_cache()

    if type(layer) is HQQLinear:
        return hqq_aoint4_layer

    if type(layer) is HQQLinearLoRA:
        layer.linear_layer = hqq_aoint4_layer

    return layer


def replace_with_torchInt4(model):
    patch_linearlayers(model, patch_hqq_to_aoint4)
    cleanup()


# Force requantize, mainly to check if the padding with int4mm is faster
def patch_hqq_to_aoint4_force_requantize(layer, patch_params):
    hqq_layer = None
    if type(layer) is HQQLinear:
        hqq_layer = layer
    if type(layer) is HQQLinearLoRA:
        hqq_layer = layer.linear_layer

    if hqq_layer is None:
        return layer

    if hqq_layer.meta["nbits"] != 4 or hqq_layer.meta["axis"] != 1:
        print("Skipping aoint4 conversion for ", hqq_layer.name)
        return layer

    # Create dummy linear layer to store dequantize weights
    dummy_linear = torch.nn.Linear(1, 1, bias=False)
    dummy_linear.weight.data = hqq_layer.dequantize()

    # Disable optimizer on already dequantized weights
    quant_config = hqq_layer.quant_config
    quant_config["weight_quant_params"]["optimize"] = False

    hqq_aoint4_layer = HQQLinearTorchWeightOnlynt4(
        dummy_linear,
        quant_config=quant_config,
        compute_dtype=hqq_layer.compute_dtype,
        device=hqq_layer.device,
        del_orig=True,
        initialize=True,
        padding=True,
    )

    del hqq_layer
    torch.cuda.empty_cache()

    if type(layer) is HQQLinear:
        return hqq_aoint4_layer

    if type(layer) is HQQLinearLoRA:
        layer.linear_layer = hqq_aoint4_layer

    return layer


def replace_with_torchInt4_force_requantize(model):
    patch_linearlayers(model, patch_hqq_to_aoint4_force_requantize)
    cleanup()



# Written by Dr. Hicham Badri @Mobius Labs GmbH - 2024
#####################################################

# Rewrite generate() to support torch.compile fullgraph.
# https://gist.github.com/ArthurZucker/5dc54a3fb443e979fac437e5df7c800b

import torch
from typing import Union
from transformers import StaticCache
from tqdm import tqdm


class HFGenerator:
    def __init__(
        self,
        model,
        tokenizer,
        max_new_tokens: int = 1000,
        cache_size: Union[int,None] = None,
        do_sample: bool = False,
        temperature: float = 0.6,
        top_k: int = 5,
        compile: Union[str,None] = None, #None / "partial" / "full"
    ):
        super().__init__()

        torch._dynamo.config.cache_size_limit = 32
        torch._dynamo.config.capture_scalar_outputs = True

        self.model = model
        self.tokenizer = tokenizer
        self.device = model.device
        self.do_sample = do_sample
        self.temperature = temperature if self.do_sample else None
        self.top_k = top_k if self.do_sample else None
        self.use_cache = False

        if do_sample:
            decode_one_token = self.decode_one_token_sampled
        else:
            decode_one_token = self.decode_one_token_no_sample

        # Setup cache
        self.max_new_tokens = max_new_tokens
        if cache_size is None:
            self.cache_size = self.next_multiple(self.max_new_tokens)
        else:
            self.cache_size = cache_size

        self.max_new_tokens = min(self.max_new_tokens, self.cache_size)

        self.setup_cache()

        self.is_compiled = False
        if compile == "partial":
            self.compile_partial(decode_one_token)

        if compile == "full":
            self.compile_full()

        if(hasattr(self, 'decode_one_token') is False):
            self.decode_one_token = decode_one_token

        self.init()

    @torch.no_grad()
    def setup_cache(self):
        self.model._setup_cache(StaticCache, 1, max_cache_len=self.cache_size)

    # Ideally only compile this, but it creates issues with generation: https://github.com/huggingface/transformers/issues/30351
    def compile_partial(self, decode_one_token):
        self.decode_one_token = torch.compile(
            decode_one_token, **{"mode": "reduce-overhead", "fullgraph": True}
        )
        self.is_compiled = True

    @torch.inference_mode()
    def compile_full(self):
        self.model.forward = torch.compile(
            self.model.forward, mode="reduce-overhead", fullgraph=True
        )

        with torch.backends.cuda.sdp_kernel(
            enable_flash=True, enable_mem_efficient=False, enable_math=True
        ):
            for ctx_size in [1] * 10:
                self.model(
                    torch.ones(
                        (1, ctx_size), dtype=torch.int32, device=self.model.device
                    ),
                    use_cache=False,
                ).logits

        self.is_compiled = True

    def next_multiple(self, val):  # next power of 2
        vals = [
            2**i for i in range(5, 14)
        ]  # [32, 64, 128, 256, 512, 1024, 2048, 4096, 8192]
        new_val = vals[[i for i in range(len(vals)) if (vals[i] - val) > 0][0]]
        return new_val

    def init(self):
        # Setup inference mode
        self.tokenizer.add_bos_token = False
        self.tokenizer.add_eos_token = False
        if not self.tokenizer.pad_token:
            self.tokenizer.add_special_tokens({"pad_token": "<<[PAD]>>"})
        self.tokenizer.padding_side = "right"
        self.model.eval()
        self.model.config.use_cache = True

    # Copied from https://gist.github.com/ArthurZucker/af34221def212259b43d55a2811d2dbb
    def multinomial_sample_one_no_sync(
        self, probs_sort
    ):  # Does multinomial sampling without a cuda synchronization
        q = torch.empty_like(probs_sort).exponential_(1)
        return torch.argmax(probs_sort / q, dim=-1, keepdim=True).to(dtype=torch.int)

    # Copied from https://gist.github.com/ArthurZucker/af34221def212259b43d55a2811d2dbb
    def logits_to_probs(self, logits, temperature=1.0, top_k=None):
        logits = logits / max(temperature, 1e-5)
        if top_k is not None:
            v, _ = torch.topk(logits, min(top_k, logits.size(-1)))
            pivot = v.select(-1, -1).unsqueeze(-1)
            logits = torch.where(logits < pivot, -float("Inf"), logits)
        probs = torch.nn.functional.softmax(logits, dim=-1)
        return probs

    # Copied from https://gist.github.com/ArthurZucker/af34221def212259b43d55a2811d2dbb
    def sample(self, logits, temperature, top_k):
        probs = self.logits_to_probs(logits[:, -1], temperature, top_k)
        idx_next = self.multinomial_sample_one_no_sync(probs)
        return idx_next, probs

    def decode_one_token_no_sample(
        self, cur_token, input_pos, cache_position, temperature=None, top_k=None
    ):
        logits = self.model(
            cur_token,
            position_ids=input_pos,
            cache_position=cache_position,
            return_dict=False,
            use_cache=self.use_cache,
        )[0]
        new_token = torch.argmax(logits[:, -1], dim=-1)[:, None]
        return new_token

    def decode_one_token_sampled(
        self, cur_token, input_pos, cache_position, temperature=0.6, top_k=5
    ):
        logits = self.model(
            cur_token,
            position_ids=input_pos,
            cache_position=cache_position,
            return_dict=False,
            use_cache=self.use_cache,
        )[0]
        new_token = self.sample(logits, temperature=temperature, top_k=top_k)[0]
        return new_token

    # Setup cache and variables
    def setup(self, inputs, max_new_tokens):
        #Temporary hack: this is why results with compiled are weird. They should be fixed https://github.com/huggingface/transformers/issues/30351
        if(self.is_compiled is False):
            self.setup_cache()
        self.inputs = inputs
        self.batch_size, self.seq_length = self.inputs["input_ids"].shape
        self.cache_position = torch.arange(self.seq_length, device=self.device)
        self.generated_ids = torch.zeros(
            self.batch_size,
            self.seq_length + max_new_tokens + 1,
            dtype=torch.int,
            device=self.device,
        )
        self.generated_ids[:, self.cache_position] = self.inputs["input_ids"].to(
            torch.int
        )

    # Pre-fill phase
    def prefill(self):
        logits = self.model(
            **self.inputs,
            cache_position=self.cache_position,
            return_dict=False,
            use_cache=self.use_cache,
        )[0]
        next_token = torch.argmax(logits[:, -1], dim=-1)[:, None]
        self.generated_ids[:, self.seq_length] = next_token[:, 0]
        self.cache_position = torch.tensor(
            [self.seq_length], device=self.device, dtype=torch.int
        )
        self.begin_gen_position = self.cache_position.item()

        return next_token

    # generate one token at a time
    def gen_next_token(self, next_token):
        with torch.backends.cuda.sdp_kernel(
            enable_flash=True, enable_mem_efficient=False, enable_math=True
        ):
            next_token = self.decode_one_token(
                next_token.clone(),
                None,
                self.cache_position + 1,
                temperature=self.temperature,
                top_k=self.top_k,
            )
        self.cache_position += 1
        self.generated_ids[:, self.cache_position] = next_token.int()
        return next_token

    def print_current_token(self, output_text_len):
        output_text = self.tokenizer.decode(
            self.generated_ids[0, self.begin_gen_position : self.cache_position + 1]
        )
        printable_text = output_text[output_text_len:]
        output_text_len = len(output_text)
        print(printable_text, end="", flush=True)
        return output_text_len

    def next_token_iterator(
        self, next_token, max_new_tokens, verbose, print_tokens, cleanup=True
    ):
        output_text, output_text_len = "", 0
        for i in tqdm(range(1, max_new_tokens), disable=(not verbose or print_tokens)):
            next_token = self.gen_next_token(next_token)

            if next_token[0].item() == self.tokenizer.eos_token_id:
                break

            # You need to keep track of the whole text, otherwise you lose spaces, makes everything much slower ¯\_(ツ)_/¯
            # https://github.com/huggingface/transformers/blob/b109257f4fb8b1166e7c53cc5418632014ed53a5/src/transformers/generation/streamers.py#L95-L114
            if print_tokens:
                output_text_len = self.print_current_token(output_text_len)

        input_tokens = self.generated_ids[0, : self.begin_gen_position].cpu()
        output_tokens = self.generated_ids[
            0, self.begin_gen_position : self.cache_position
        ].cpu()
        output_text = self.tokenizer.decode(output_tokens)

        if cleanup:
            # model._reset_cache()
            del self.inputs, self.generated_ids, self.cache_position
            torch.cuda.empty_cache()

        return {
            "output_text": output_text,
            "output_tokens": output_tokens,
            "input_tokens": input_tokens,
        }

    @torch.inference_mode()
    def generate(
        self, prompt, use_chat_template=True, verbose=True, print_tokens=False
    ):
        self.setup(
            inputs=self.tokenize_prompt(prompt, use_chat_template=use_chat_template),
            max_new_tokens=self.max_new_tokens,
        )
        return self.next_token_iterator(
            self.prefill(), self.max_new_tokens, verbose, print_tokens
        )

    def generate_(
        self, prompt, use_chat_template=True, verbose=False, print_tokens=False
    ):
        gen_out = self.model.generate(
            **self.tokenize_prompt(prompt, use_chat_template=use_chat_template),
            do_sample=self.do_sample,
            cache_implementation="static",
            max_new_tokens=self.max_new_tokens,
            pad_token_id=self.tokenizer.pad_token_id,
            temperature=self.temperature,
            top_p=self.top_k,
            use_cache=False,
        )[0]

        return {"output_text": self.tokenizer.decode(gen_out), "output_tokens": gen_out}

    def tokenize_prompt(self, prompt, use_chat_template=True):
        if use_chat_template:
            self.tokenizer.add_bos_token = True
            prompt = self.tokenizer.apply_chat_template(
                [
                    {"role": "user", "content": prompt},
                ],
                tokenize=False,
            )

        self.tokenizer.add_bos_token = False

        return self.tokenizer([prompt], return_tensors="pt", padding=True).to(
            device=self.model.device
        )



# Written by Dr. Hicham Badri @Mobius Labs GmbH - 2024
#####################################################
import torch
from torch import Tensor
from ..core.quantize import Quantizer, HQQLinear
from ..core.utils import cleanup
from ..core.peft import HQQLinearLoRA
from ..models.hf.base import AutoHQQHFModel
from ..backends.torchao import patch_hqq_to_aoint4
from ..backends.marlin import patch_hqq_to_marlin


def patch_linearlayers(model, fct, patch_param=None, verbse=False):
    base_class = model.base_class if (hasattr(model, "base_class")) else AutoHQQHFModel
    base_class.setup_model(model)
    model.base_class.patch_linearlayers(
        model, fct, dict([(k, patch_param) for k in model.linear_tags]), verbose=verbse
    )


def patch_add_quant_config(layer, patch_param):
    if type(layer) is HQQLinear:
        layer.quant_config = patch_param
    if type(layer) is HQQLinearLoRA:
        layer.linear_layer.quant_config = patch_param
    return layer


# add dummy weights to a layer
def patch_add_weight_param(layer, patch_param):
    if hasattr(layer, "weight") is False:
        if hasattr(layer, "device"):
            device_ = layer.device
        else:
            param = [p for p in layer.parameters()]
            device_ = param[0].device if (len(param) > 0) else patch_param["device"]

        fp_param = [p for p in layer.parameters() if p.is_floating_point()]
        dtype_ = fp_param[0].dtype if (len(fp_param) > 0) else patch_param["dtype"]

        layer.weight = torch.nn.Parameter(
            torch.zeros((1,), device=device_, dtype=dtype_), requires_grad=False
        )
    return layer


# Optimize HQQLinear.forward for inference
def patch_hqq_inference(layer, patch_param):
    def forward_hqq_inferece(self, x):
        out = torch.matmul(x, self.dequantize().T)  # TODO GEMV use-case
        if self.bias is not None:
            out += self.bias
        return out

    if type(layer) is HQQLinear:
        layer.forward = lambda x: forward_hqq_inferece(layer, x)

    if type(layer) is HQQLinearLoRA:
        if type(layer.linear_layer) is HQQLinear:
            layer.linear_layer.forward = lambda x: forward_hqq_inferece(
                layer.linear_layer, x
            )

    return layer


# Optimize HQQLinearLoRA.forward for inference
def patch_lora_inference(layer, patch_param):
    def forward_lora_inference(self, x):
        out = torch.matmul(torch.matmul(x, self.lora_A), self.lora_B) * self.scaling
        return out

    if type(layer) is HQQLinearLoRA:
        layer.forward_lora = lambda x: forward_lora_inference(layer, x)
    return layer


def prepare_for_inference(model, allow_merge=False, backend="default"):
    if backend == "torchao_int4":
        allow_merge = False

    patch_linearlayers(model, patch_hqq_inference)
    patch_linearlayers(model, patch_lora_inference)
    cleanup()

    if backend == "torchao_int4":
        patch_linearlayers(model, patch_hqq_to_aoint4)
        cleanup()
    if allow_merge:  # only compatible with symmetric quant kernels
        patch_linearlayers(
            model, patch_merge_zeros_with_lora, {"z_shift": 8, "keep_lora": False}
        )
        cleanup()
    if backend == "marlin":
        patch_linearlayers(model, patch_hqq_to_marlin, verbse=True)
        cleanup()

    patch_linearlayers(
        model, patch_add_weight_param, {"device": model.device, "dtype": model.dtype}
    )
    cleanup()


def get_lowrank_tuple_torch_gpu(tensor, max_rank, eps=None):
    t = tensor.t().float()
    u, s, v = torch.linalg.svd(t)
    u, s, v = u[:, :max_rank], s[:max_rank], v[:max_rank, :]
    us = torch.matmul(u, torch.diag(s))
    A, B = (v.t(), us.t())  # t ~ AB
    if eps is not None:
        A = A.clamp(min=-eps, max=eps)
        B = B.clamp(min=-eps, max=eps)
    return A.to(tensor.dtype), B.to(tensor.dtype)


# Merges HQQ's zeros with LoRA weights. ONLY works with group_size=None and axis=1
def patch_merge_zeros_with_lora(layer, patch_params={"z_shift": 8, "keep_lora": False}):
    if type(layer) is HQQLinearLoRA:
        # Check config suppport
        hqq_layer = layer.linear_layer
        if (hqq_layer.meta["axis"] == 0) or (hqq_layer.meta["group_size"] is not None):
            print('Skipping zeros lora merging for', layer.name)
            return layer

        layer.z_shift = patch_params["z_shift"]
        layer.keep_lora = patch_params["keep_lora"]

        shape = layer.linear_layer.meta["shape"]
        z = layer.linear_layer.meta["zero"]
        s = layer.linear_layer.meta["scale"]
        u = (s * (-z + layer.z_shift)).flatten().view([1, -1])

        ###########################################
        if layer.keep_lora is False:
            A, B = layer.lora_A.data, layer.lora_B.data
            onz = torch.ones((shape[1], 1), device=u.device, dtype=u.dtype)

            # Cat
            A = torch.cat([A, onz], axis=1)
            B = torch.cat([B, u], axis=0)

            # #Re-rank
            # #A, B = get_lowrank_tuple_torch_gpu(torch.matmul(A, B) + torch.matmul(onz, u), max_rank=layer.r + 1)

            layer.lora_A.data = A.to(dtype=layer.lora_A.dtype)
            layer.lora_B.data = B.to(dtype=layer.lora_B.dtype)

            layer.u = None

        else:
            layer.u = torch.nn.Parameter(u, requires_grad=False)
        ###########################################
        layer.linear_layer.meta["zero"] = 0.0

        torch.cuda.empty_cache()

        def forward_linear_updated(self, x: Tensor) -> Tensor:
            compute_dtype = self.linear_layer.compute_dtype
            meta = self.linear_layer.meta
            W_q = self.linear_layer.W_q

            W_r = Quantizer.unpack[meta["packing"]](W_q, dtype=compute_dtype).t()
            scale = meta["scale"].t()

            out = torch.matmul(x, (W_r - self.z_shift)) * scale  # Symmetric quant
            return out

        layer.linear_layer.forward = lambda x: forward_linear_updated(layer, x)

        def forward_updated(self, x: Tensor) -> Tensor:
            out = self.linear_layer.forward(x)

            if self.u is not None:
                out += torch.matmul(x.sum(axis=-1, keepdim=True), self.u)

            out += self.forward_lora(x) + self.bias
            return out

        layer.forward = lambda x: forward_updated(layer, x)

    return layer



# Written by Dr. Hicham Badri @Mobius Labs GmbH - 2023
#####################################################
import torch
from torch import uint8, int32, float16, nn, Tensor
import copy
from enum import Enum
from typing import Union

from .utils import is_divisible
from .optimize import optimize_weights_proximal
from .bitpack import BitPack


# Main HQQ Quantizer
class Quantizer:
    SUPPORTED_BITS = [8, 6, 5, 4, 3, 2, 1]
    optimize_weights = optimize_weights_proximal

    bit_to_packing = {
        8: "8bit_u8",
        6: "8bit_u8", #todo: bitpacking
        5: "8bit_u8", #todo: bitpacking
        4: "4bit_u8",
        3: "3bit_32",
        2: "2bit_u8",
        1: "1bit_u8",
    }

    pack = {
        "8bit_u8": BitPack.pack_8bit_u8,
        "4bit_u8": BitPack.pack_4bit_u8,
        "3bit_32": BitPack.pack_3bit_32,
        "2bit_u8": BitPack.pack_2bit_u8,
        "1bit_u8": BitPack.pack_1bit_u8,
    }

    unpack = {
        "8bit_u8": BitPack.unpack_8bit_u8,
        "4bit_u8": BitPack.unpack_4bit_u8,
        "3bit_32": BitPack.unpack_3bit_32,
        "2bit_u8": BitPack.unpack_2bit_u8,
        "1bit_u8": BitPack.unpack_1bit_u8,
    }

    unpack_view_dtype = {
        "8bit_u8": uint8,
        "4bit_u8": uint8,
        "3bit_32": int32,
        "2bit_u8": uint8,
        "1bit_u8": uint8,
    }

    @classmethod
    def quantize(
        cls,
        tensor: Tensor,
        nbits: int = 4,
        channel_wise: bool = True,
        group_size: int = 64,
        optimize: bool = False,
        round_zero: bool = False,
        axis: int = 0,
        bitpack: bool = True,
        compute_dtype: Union[torch.dtype, None] = None,
        view_as_float: bool = False,
        device: str = "cuda",
    ) -> tuple:
        assert nbits in Quantizer.SUPPORTED_BITS, (
            "nbits=" + str(nbits) + " not supported."
        )
        assert axis in [0, 1], "axis should be either 0 or 1"
        if group_size is not None:
            assert is_divisible(tensor.numel(), group_size), (
                "group_size should be divisble by the total tensor dimensions. shape: "
                + str(tensor.shape)
                + ", group_size: "
                + str(group_size)
            )

        W = tensor.float()
        shape = W.shape

        # Reshape for grouping
        if (group_size is not None) and channel_wise:
            W = (
                W.reshape([-1, group_size])
                if (axis == 1)
                else W.reshape([group_size, -1])
            )

        # Get min/max values
        if not channel_wise:
            _min, _max = W.min(), W.max()
            optimize = False
        else:
            _min = W.min(axis=axis, keepdim=True)[0]
            _max = W.max(axis=axis, keepdim=True)[0]

        max_v = 2**nbits - 1
        min_v = 0
        min_max = [min_v, max_v]

        # Note: here we work with the inverse of the scale to avoid division and quantize instead via W*scale + zero, the scale is inverted later on.
        scale = (max_v / (_max - _min)).clamp(
            max=2e4
        )  # clamp to avoid half-precision problems
        zero = -_min * scale

        # Round zero as in: https://github.com/casper-hansen/AutoAWQ/blob/main/awq/quantize/quantizer.py#L42C9-L42C14
        if round_zero:
            zero = torch.round(zero)

        # Fine-tune weights
        if optimize:
            W_q, scale, zero = Quantizer.optimize_weights(
                tensor=W,
                scale=scale,
                zero=zero,
                min_max=min_max,
                axis=axis,
                device=device,
            )
        else:
            W_q = torch.round(W * scale + zero).clamp(min_max[0], min_max[1])

        # Store meta-data (we invert the scale for dequantization)
        meta = {
            "nbits": nbits,
            "group_size": group_size,
            "shape": shape,
            "scale": 1.0 / scale,
            "zero": zero,
            "axis": axis,
            "packing": Quantizer.bit_to_packing[nbits],
        }
        meta["unpack_view_dtype"] = Quantizer.unpack_view_dtype[meta["packing"]]

        # Pack bits
        meta["view_as_float"] = view_as_float
        if bitpack:
            W_q = Quantizer.pack[meta["packing"]](W_q)
            if view_as_float:
                W_q = W_q.view(
                    torch.float32 if compute_dtype is None else compute_dtype
                )  # store quantized weights as compute_dtype
        else:
            W_q = W_q.to(tensor.dtype)
            meta["packing"] = None

        # cleanup
        del W, _min, _max
        torch.cuda.empty_cache()

        return W_q, meta

    # Main dequantization: bit_unpacking > (W_q - z)*s > reshape
    @classmethod
    def dequantize(cls, W_q: Tensor, meta: dict) -> Tensor:
        compute_dtype = meta["compute_dtype"] if ("compute_dtype" in meta) else float16
        if meta["packing"]:
            if meta["view_as_float"]:
                W_q = W_q.view(meta["unpack_view_dtype"])
            W_r = Quantizer.unpack[meta["packing"]](W_q, dtype=compute_dtype)
            if meta["nbits"] == 3:
                W_r = W_r[
                    : meta["group_size"]
                    if meta["axis"] == 0
                    else meta["shape"][0] * meta["shape"][1] // meta["group_size"]
                ]
        else:
            W_r = W_q.to(compute_dtype)
        W_r = ((W_r - meta["zero"]) * meta["scale"]).reshape(meta["shape"])
        return W_r

    @classmethod
    def to_inplace(cls, W_q: Tensor, meta: dict, device) -> tuple:
        compute_dtype = meta["compute_dtype"] if ("compute_dtype" in meta) else float16
        if W_q is not None:
            W_q = W_q.to(device).contiguous()
        for key in meta:
            if type(meta[key]) == torch.Tensor:
                meta[key] = (
                    (
                        meta[key].to(compute_dtype)
                        if torch.is_floating_point(meta[key])
                        else meta[key]
                    )
                    .to(device)
                    .contiguous()
                )
        return W_q, meta

    @classmethod
    def to_ooplace(cls, W_q: Tensor, meta: dict, device) -> tuple:
        compute_dtype = meta["compute_dtype"] if ("compute_dtype" in meta) else float16
        if W_q is not None:
            W_q_c = W_q.to(device).contiguous()
        else:
            W_q_c = None
        meta_c = {}
        for key in meta:
            if type(meta[key]) == torch.Tensor:
                meta_c[key] = (
                    (
                        meta[key].to(compute_dtype)
                        if torch.is_floating_point(meta[key])
                        else meta[key]
                    )
                    .to(device)
                    .contiguous()
                )
            else:
                meta_c[key] = meta[key]
        return W_q_c, meta_c

    @classmethod
    def cuda(cls, W_q: Tensor, meta: dict, device) -> tuple:
        return Quantizer.to_inplace(W_q, meta, device=device)

    @classmethod
    def cpu(cls, W_q: Tensor, meta: dict) -> tuple:
        return Quantizer.to_ooplace(W_q, meta, device="cpu")


# Check Aten/CUDA backend if available
hqq_aten_is_available = False
try:
    import hqq_aten

    hqq_aten_is_available = True
except Exception:
    hqq_aten = None
    hqq_aten_is_available = False


class HQQBackend(Enum):
    # Name of the forward functions
    PYTORCH = "forward_pytorch_backprop"
    PYTORCH_COMPILE = "forward_pytorch_backprop_compile"
    ATEN = "forward_aten_backprop"

    # Alias for backward compatibility
    PYTORCH_BACKPROP = "forward_pytorch_backprop"
    PYTORCH_BACKPROP_COMPILE = "forward_pytorch_backprop_compile"
    ATEN_BACKPROP = "forward_aten_backprop"

    PYTORCH_FORWARD = "forward_pytorch"
    PYTORCH_FORWARD_COMPILE = "forward_pytorch_compile"
    ATEN_FORWARD = "forward_aten"

    # Experimental
    ATEN_FORWARD_INT8 = "forward_aten_int8"


# No cache: less memory, slower
class HQQMatmulNoCacheDeq(torch.autograd.Function):
    @staticmethod
    def forward(x: Tensor, dequantize, bias: Tensor):
        out = torch.matmul(x, dequantize().t())
        if bias is not None:
            out += bias
        return out

    @staticmethod
    def setup_context(ctx, inputs, outputs):
        x, dequantize, bias = inputs
        ctx.save_for_backward(x, bias)
        ctx.dequantize = dequantize

    @staticmethod
    def backward(ctx, grad_output):
        x, bias = ctx.saved_tensors

        grad_input = grad_weight = grad_bias = None

        if ctx.needs_input_grad[0]:
            grad_input = torch.matmul(grad_output, ctx.dequantize())

        # weight grad for frozen quantized weights not defined
        # if ctx.needs_input_grad[1]:
        #   grad_weight = torch.matmul(grad_output.t(), x)

        if bias is not None and ctx.needs_input_grad[2]:
            grad_bias = grad_output.sum(0)

        return grad_input, grad_weight, grad_bias


class HQQMatmulNoCacheMul(torch.autograd.Function):
    @staticmethod
    def forward(x, matmul, bias):
        out = matmul(x, transpose=True)
        if bias is not None:
            out += bias
        return out

    @staticmethod
    def setup_context(ctx, inputs, outputs):
        x, matmul, bias = inputs
        ctx.save_for_backward(x, bias)
        ctx.matmul = matmul

    @staticmethod
    def backward(ctx, grad_output):
        x, bias = ctx.saved_tensors

        grad_input = grad_weight = grad_bias = None

        if ctx.needs_input_grad[0]:
            grad_input = ctx.matmul(grad_output, transpose=False)

        # weight grad for frozen quantized weights not defined
        # if ctx.needs_input_grad[1]:
        #   grad_weight = torch.matmul(grad_output.t(), x)

        if bias is not None and ctx.needs_input_grad[2]:
            grad_bias = grad_output.sum(0)

        return grad_input, grad_weight, grad_bias


# Cache dequantized tensor: Faster but needs more memory
class HQQMatmulCachedDeq(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x, hqq_layer, bias):
        weight_tmp = hqq_layer.dequantize()
        out = torch.matmul(x, weight_tmp.t())
        if bias is not None:
            out += bias

        ctx.save_for_backward(x, bias, weight_tmp)
        return out

    @staticmethod
    def backward(ctx, grad_output):
        x, bias, weight_tmp = ctx.saved_tensors

        grad_input = grad_weight = grad_bias = None

        if ctx.needs_input_grad[0]:
            grad_input = torch.matmul(grad_output, weight_tmp)

        del weight_tmp

        if bias is not None and ctx.needs_input_grad[2]:
            grad_bias = grad_output.sum(0)

        return grad_input, grad_weight, grad_bias


# Main linear layer
class HQQLinear(nn.Module):
    # Default backend
    backend = HQQBackend.PYTORCH

    def __init__(
        self,
        linear_layer: Union[nn.Module, None],
        quant_config: dict,
        del_orig: bool = True,
        compute_dtype: torch.dtype = float16,
        device: str = "cuda",
        initialize: bool = True,
    ):
        super().__init__()
        self.ready = False
        self.in_gpu = False
        self.bias = None
        self.device = device
        self.compute_dtype = compute_dtype
        self.quant_config = copy.deepcopy(quant_config)
        self.del_orig = del_orig
        self.offload_meta = (
            self.quant_config.pop("offload_meta")
            if (self.quant_config is not None)
            else None
        )

        self.set_backend(HQQLinear.backend)

        self.linear_layer = linear_layer
        self.W_q = None
        self.meta = None

        if initialize:
            self.initialize()

    def initialize(self):
        if self.linear_layer is not None:
            self.quantize(self.linear_layer.weight.data, **self.quant_config)
            self.bias = (
                None
                if (self.linear_layer.bias is None)
                else self.linear_layer.bias.to(
                    device=self.device, dtype=self.compute_dtype
                )
            )

        if self.del_orig:
            del self.linear_layer
        torch.cuda.empty_cache()

    def extra_repr(self) -> str:
        out = ""
        if hasattr(self, "meta"):
            if self.meta is not None:
                in_features, out_features = self.meta["shape"][::-1]
                out = f"in_features={in_features}, out_features={out_features}, bias={self.bias is not None}"
        return out

    # Set backends
    @classmethod
    def set_backend(cls, backend: HQQBackend):
        if "aten" in backend.value:
            if hqq_aten_is_available is False:
                print(
                    "ATEN/CUDA backend not availabe. Make sure you install the hqq_aten library."
                )
                return
            print(
                "Warning: the ATEN/CUDA backend only supports axis=0 and GPU runtime."
            )
        HQQLinear.backend = backend
        cls.forward = getattr(cls, backend.value)

    # TODO: rewrite this mess
    def cuda(self, device):
        self.meta["compute_dtype"] = self.compute_dtype

        if type(self.W_q) == nn.parameter.Parameter:
            self.W_q.data, self.meta = Quantizer.cuda(self.W_q.data, self.meta, device)
        else:
            self.W_q, self.meta = Quantizer.cuda(self.W_q, self.meta, device)

        if self.meta["quant_zero"]:
            if "zero_q" in self.meta:
                self.meta["zero_q"], self.meta["meta_zero"] = Quantizer.cuda(
                    self.meta["zero_q"], self.meta["meta_zero"], device
                )
            else:
                _, self.meta["meta_zero"] = Quantizer.cuda(
                    None, self.meta["meta_zero"], device
                )
        else:
            self.meta["zero"] = self.meta["zero"].to(device)

        if self.meta["quant_scale"]:
            if "scale_q" in self.meta:
                self.meta["scale_q"], self.meta["meta_scale"] = Quantizer.cuda(
                    self.meta["scale_q"], self.meta["meta_scale"], device
                )
            else:
                _, self.meta["meta_scale"] = Quantizer.cuda(
                    None, self.meta["meta_scale"], device
                )
        else:
            self.meta["scale"] = self.meta["scale"].to(device)

        # #Use zero/scale with streams for dequantization is faster than packing in "zero_scale"
        # for key in ["zero", "zero_q", "scale", "scale_q"]:
        #     if((key in self.meta) and self.offload_meta):
        #         self.meta[key] = self.meta[key].contiguous().cpu().pin_memory()

        if self.offload_meta:
            if "zero_scale" not in self.meta:
                if self.meta["quant_scale"] and self.meta["quant_zero"]:
                    self.meta["zero_scale"] = torch.stack(
                        (self.meta["zero_q"], self.meta["scale_q"])
                    )
                    del self.meta["scale_q"], self.meta["zero_q"]
                else:
                    self.meta["zero_scale"] = torch.stack(
                        (self.meta["zero"], self.meta["scale"])
                    ).to(self.compute_dtype)
                    del self.meta["scale"], self.meta["zero"]

            self.meta["zero_scale"] = (
                self.meta["zero_scale"].contiguous().cpu().pin_memory()
            )

        if self.bias is not None:
            self.bias = self.bias.to(device=device, dtype=self.compute_dtype)

        self.W_q = nn.Parameter(self.W_q, requires_grad=False)
        self.device = device
        self.in_gpu = True

        torch.cuda.empty_cache()

        return self

    def to(self, *args, **kwargs):
        # TODO: later
        return self

    # TODO: later
    # def to_empty(self, device, recurse=True):
    #     return self.cuda(device)

    def type(self, dst_type):
        # TODO: later
        return self

    def half(self, *args, **kwargs):
        return self

    def bfloat16(self, *args, **kwargs):
        # TODO: later
        return self

    def float(self, *args, **kwargs):
        # TODO: later
        return self

    def double(self, *args, **kwargs):
        return self

    def cpu(self):
        # TODO: later
        return self

    def state_dict(self, *args, **kwargs):  # nn.Module override compatible
        state = {"W_q": self.W_q, "meta": self.meta, "bias": self.bias}
        if "destination" in kwargs and "prefix" in kwargs:
            for key, value in state.items():
                kwargs["destination"][kwargs["prefix"] + key] = value
        return state

    def _load_from_state_dict(
        self,
        state_dict,
        prefix,
        local_metadata,
        strict,
        missing_keys,
        unexpected_keys,
        error_msgs,
    ):
        W_q_key = prefix + "W_q"
        meta_key = prefix + "meta"
        bias_key = prefix + "bias"

        if W_q_key not in state_dict:
            missing_keys.append(W_q_key)
        if meta_key not in state_dict:
            missing_keys.append(meta_key)
        if missing_keys:
            return  # Can't load weights if either weight or meta is missing

        W_q = nn.Parameter(state_dict.pop(W_q_key), requires_grad=False)
        meta = state_dict.pop(meta_key)
        bias = state_dict.pop(bias_key, None)

        unexpected_keys += state_dict.keys()

        self.load_state_dict({"W_q": W_q, "meta": meta, "bias": bias}, strict)

    def load_state_dict(self, state_dict, strict=True, assign=False):
        self.W_q = state_dict["W_q"]
        self.meta = state_dict["meta"]
        self.bias = state_dict["bias"] if ("bias" in state_dict) else None

        # Meta-data offloading
        self.offload_meta = False
        for key in ["zero", "zero_q", "scale", "scale_q", "zero_scale"]:
            if key in self.meta:
                if self.meta[key].device.type == "cpu":
                    self.offload_meta = True
                    self.meta[key] = self.meta[key].contiguous().pin_memory()

        # Float view settings
        if "unpack_view_dtype" not in self.meta:
            self.meta["unpack_view_dtype"] = Quantizer.unpack_view_dtype[
                self.meta["packing"]
            ]

        if "view_as_float" not in self.meta:
            self.meta["view_as_float"] = False

        if "meta_scale" in self.meta:
            if "view_as_float" not in self.meta["meta_scale"]:
                self.meta["meta_scale"]["view_as_float"] = False

        if "meta_zero" in self.meta:
            if "view_as_float" not in self.meta["meta_zero"]:
                self.meta["meta_zero"]["view_as_float"] = False

        # Check GPU
        self.cuda(self.device)
        self.ready = True

        # Set in_features/out_features
        self.in_features, self.out_features = self.meta["shape"][::-1]

    def quantize(
        self,
        W: Tensor,
        weight_quant_params: dict,
        scale_quant_params: dict,
        zero_quant_params: dict,
    ) -> None:
        quant_scale = scale_quant_params is not None
        quant_zero = zero_quant_params is not None

        self.in_features, self.out_features = W.t().shape

        # Quantize
        W_q, meta = Quantizer.quantize(
            W,
            device=self.device,
            compute_dtype=self.compute_dtype,
            **weight_quant_params,
        )
        meta.update({"quant_scale": quant_scale, "quant_zero": quant_zero})

        if meta["quant_zero"]:
            meta["zero_q"], meta["meta_zero"] = Quantizer.quantize(
                meta["zero"],
                device=self.device,
                view_as_float=False,
                **zero_quant_params,
            )
            del meta["zero"]
            meta["meta_zero"]["compute_dtype"] = self.compute_dtype

        if meta["quant_scale"]:
            meta["scale_q"], meta["meta_scale"] = Quantizer.quantize(
                meta["scale"],
                device=self.device,
                view_as_float=False,
                **scale_quant_params,
            )
            del meta["scale"]
            meta["meta_scale"]["compute_dtype"] = self.compute_dtype

        self.W_q = W_q
        self.meta = meta
        self.cuda(self.device)
        self.ready = True

    def dequantize(self):
        assert self.ready, "model was not quantized"
        W_q, meta = self.W_q, self.meta
        device = W_q.device
        del_keys = set()

        # Zero/Scale packed together
        if "zero_scale" in meta:
            zero_scale = meta["zero_scale"].to(device=device)

            if zero_scale.dtype == uint8:
                meta["zero_q"], meta["scale_q"] = zero_scale[0], zero_scale[1]
                del_keys.update({"zero_q", "scale_q"})
            else:
                meta["zero"], meta["scale"] = zero_scale[0], zero_scale[1]
                del_keys.update({"zero", "scale"})

        if meta["quant_zero"]:
            meta["zero"] = Quantizer.dequantize(
                meta["zero_q"].to(device=device), meta["meta_zero"]
            )
            del_keys.add("zero")

        if meta["quant_scale"]:
            meta["scale"] = Quantizer.dequantize(
                meta["scale_q"].to(device=device), meta["meta_scale"]
            )
            del_keys.add("scale")

        W_est = Quantizer.dequantize(W_q, meta)

        # Cleanup
        for key in del_keys:
            del meta[key]
        return W_est

    def matmul(self, x: Tensor, transpose: bool = True) -> Tensor:
        weight = self.dequantize()
        return torch.matmul(x, weight.t() if (transpose) else weight)

    @torch.compile()
    def matmul_compile(self, *args, **kwargs):
        return self.matmul(*args, **kwargs)

    def forward_pytorch_backprop(self, x: Tensor) -> Tensor:
        return HQQMatmulNoCacheMul.apply(x, self.matmul, self.bias)

    def forward_pytorch_backprop_compile(self, x: Tensor) -> Tensor:
        return HQQMatmulNoCacheMul.apply(x, self.matmul_compile, self.bias)

    def forward_pytorch(self, x: Tensor) -> Tensor:
        out = torch.matmul(x, self.dequantize().t())
        if self.bias is not None:
            out += self.bias
        return out

    @torch.compile()
    def forward_pytorch_compile(self, x: Tensor) -> Tensor:
        return self.forward_pytorch(x)

    ############################################################################################
    # ATen C++ / CUDA Bacekdn
    ##########################################################################################
    # Requires building the aten backend
    @torch.jit.ignore
    def dequantize_Wq_aten(self, W_q: Tensor, meta: dict):
        if meta["view_as_float"]:
            W_q = W_q.view(meta["unpack_view_dtype"])

        return hqq_aten.dequantize(
            W_q,
            meta["scale"],
            meta["zero"],
            meta["shape"],
            meta["group_size"] if (meta["group_size"]) else -1,
            meta["nbits"],
            meta["axis"],
            meta["packing"],
        )

    def dequantize_aten(self):
        # Dequantize
        assert self.ready, "model was not quantized"
        assert (
            self.meta["axis"] == 0
        ), "only axis=0 is supported. Use HQQLinear.set_backend(HQQBackend.PYTORCH) instead."

        W_q, meta = self.W_q, self.meta
        device = W_q.device
        del_keys = set()

        # Zero/Scale packed together
        if "zero_scale" in meta:
            zero_scale = meta["zero_scale"].to(device=device, non_blocking=True)
            if zero_scale.dtype == uint8:
                meta["zero_q"], meta["scale_q"] = zero_scale[0], zero_scale[1]
                del_keys.update({"zero_q", "scale_q"})
            else:
                meta["zero"], meta["scale"] = zero_scale[0], zero_scale[1]
                del_keys.update({"zero", "scale"})

        # Dequantize zero_q / scale_q with device loading
        if meta["quant_zero"]:
            if meta["meta_zero"]["group_size"]:
                meta["zero"] = self.dequantize_Wq_aten(
                    meta["zero_q"].to(device=device), meta["meta_zero"]
                )
            else:
                meta["zero"] = Quantizer.dequantize(
                    meta["zero_q"].to(device=device), meta["meta_zero"]
                )
            del_keys.add("zero")

        if meta["quant_scale"]:
            if meta["meta_scale"]["group_size"]:
                meta["scale"] = self.dequantize_Wq_aten(
                    meta["scale_q"].to(device=device), meta["meta_scale"]
                )
            else:
                meta["scale"] = Quantizer.dequantize(
                    meta["scale_q"].to(device=device), meta["meta_scale"]
                )
            del_keys.add("scale")

        # Reconstruct the weights
        W_est = self.dequantize_Wq_aten(W_q, meta)

        # Cleanup
        for key in del_keys:
            del meta[key]

        return W_est

    # Much faster with data-offloading zero_q/scale_q but takes more VRAM
    def dequantize_aten_with_streams(self):
        # Create streams
        if hasattr(self, "stream_zero") is False:
            self.stream_zero = torch.cuda.Stream()
            self.stream_scale = torch.cuda.Stream()

        # Dequantize
        assert self.ready, "model was not quantized"
        W_q, meta = self.W_q, self.meta
        device = W_q.device
        del_keys = set()

        # Zero/Scale packed together
        if "zero_scale" in meta:
            zero_scale = meta["zero_scale"].to(device=device, non_blocking=True)
            if zero_scale.dtype == uint8:
                meta["zero_q"], meta["scale_q"] = zero_scale[0], zero_scale[1]
                del_keys.update({"zero_q", "scale_q"})
            else:
                meta["zero"], meta["scale"] = zero_scale[0], zero_scale[1]
                del_keys.update({"zero", "scale"})

        # Using non_blocking=False for the moment, otherwise it can result in strange behaviors
        non_blocking = False
        with torch.cuda.stream(self.stream_zero):
            if meta["quant_zero"]:
                if meta["meta_zero"]["group_size"]:
                    meta["zero"] = self.dequantize_Wq_aten(
                        meta["zero_q"].to(device=device, non_blocking=non_blocking),
                        meta["meta_zero"],
                    )
                else:
                    meta["zero"] = Quantizer.dequantize(
                        meta["zero_q"].to(device=device, non_blocking=non_blocking),
                        meta["meta_zero"],
                    )
                del_keys.add("zero")

        with torch.cuda.stream(self.stream_scale):
            if meta["quant_scale"]:
                if meta["meta_scale"]["group_size"]:
                    meta["scale"] = self.dequantize_Wq_aten(
                        meta["scale_q"].to(device=device, non_blocking=non_blocking),
                        meta["meta_scale"],
                    )
                else:
                    meta["scale"] = Quantizer.dequantize(
                        meta["scale_q"].to(device=device, non_blocking=non_blocking),
                        meta["meta_scale"],
                    )
                del_keys.add("scale")

        # Wait for streams to finish
        torch.cuda.synchronize()

        # Reconstruct the weights
        W_est = self.dequantize_Wq_aten(W_q, meta)

        # Cleanup
        for key in del_keys:
            del meta[key]

        return W_est

    def forward_aten(self, x: Tensor) -> Tensor:
        W_est = self.dequantize_aten()
        out = torch.matmul(x, W_est.t())
        if self.bias is not None:
            out += self.bias

        return out

    def forward_aten_backprop(self, x: Tensor) -> Tensor:
        return HQQMatmulNoCacheDeq.apply(x, self.dequantize_aten, self.bias)

    # TODO: as fused kernel in CUDA
    def _get_int8_matrix(self, M):
        scale = torch.abs(M).amax() / 127.0
        return torch.round(M / scale).to(torch.int8), scale.float()

    # TODO: in ATEN
    @torch.compile()
    def _matmul_int8(self, A, B):
        dtype = A.dtype
        A, scale_A = self._get_int8_matrix(A)
        B, scale_B = self._get_int8_matrix(B)
        return (torch._int_mm(A, B) * (scale_A * scale_B)).to(dtype)

    def forward_aten_int8(self, x: Tensor) -> Tensor:
        W_est = self.dequantize_aten()
        out = self._matmul_int8(x[0], W_est.t())[None, ...]
        if self.bias is not None:
            out += self.bias

        return out


def hqq_base_quant_config(
    nbits: int = 4,
    group_size: int = 64,
    quant_zero: bool = True,
    quant_scale: bool = False,
    offload_meta: bool = False,  # meta-data should be quantized with the same settings to use offload_meta
    view_as_float: bool = False,
    axis: int = 0,
):
    assert (
        nbits in Quantizer.SUPPORTED_BITS
    ), "nbits value not supported. Check Quantizer.SUPPORTED_BITS."
    if group_size is not None:
        assert is_divisible(
            group_size, 8
        ), "Invalid group_size param: the value should be a multiple of 8."
    weight_quant_params = {
        "nbits": nbits,
        "channel_wise": True,
        "group_size": group_size,
        "optimize": True,
        "round_zero": True if nbits == 4 else False,
        "axis": axis,
        "view_as_float": view_as_float,
    }

    if offload_meta:
        if quant_scale != quant_zero:
            # print(colored("quant_zero and quant_scale must be the same when offload_meta is set to True. Setting quant_scale=quant_zero." , 'yellow'))
            quant_scale = quant_zero

        scale_quant_params = (
            {"nbits": 8, "channel_wise": True, "group_size": 128, "optimize": False}
            if (quant_scale)
            else None
        )
        zero_quant_params = (
            {"nbits": 8, "channel_wise": True, "group_size": 128, "optimize": False}
            if (quant_zero)
            else None
        )

    else:
        scale_quant_params = (
            {"nbits": 8, "channel_wise": True, "group_size": 128, "optimize": False}
            if (quant_scale)
            else None
        )
        zero_quant_params = (
            {"nbits": 8, "channel_wise": False, "group_size": None, "optimize": False}
            if (quant_zero)
            else None
        )

    return {
        "weight_quant_params": weight_quant_params,
        "scale_quant_params": scale_quant_params,
        "zero_quant_params": zero_quant_params,
        "offload_meta": offload_meta,
    }


# Alias: follow similar Auto-GPTQ naming
BaseQuantizeConfig = hqq_base_quant_config



# Written by Dr. Hicham Badri @Mobius Labs GmbH - 2023
#####################################################
import timm
import json
import torch
from .base import HQQWrapper

from ..models.base import BaseHQQModel
from ..models.timm.vit_clip import ViTCLIPHQQ

_HQQ_REGISTRY = {}
_HQQ_REGISTRY["vit_huge_patch14_clip_336"] = ViTCLIPHQQ
_HQQ_REGISTRY["vit_huge_patch14_clip_224"] = ViTCLIPHQQ
_HQQ_REGISTRY["vit_large_patch14_clip_224"] = ViTCLIPHQQ
_HQQ_REGISTRY["vit_large_patch14_clip_336"] = ViTCLIPHQQ
_HQQ_REGISTRY["vit_base_patch16_clip_384"] = ViTCLIPHQQ
_HQQ_REGISTRY["vit_base_patch32_clip_448"] = ViTCLIPHQQ
_HQQ_REGISTRY["vit_base_patch32_clip_384"] = ViTCLIPHQQ
_HQQ_REGISTRY["vit_base_patch16_clip_224"] = ViTCLIPHQQ
_HQQ_REGISTRY["vit_base_patch32_clip_224"] = ViTCLIPHQQ


class HQQtimm(HQQWrapper):
    _HQQ_REGISTRY = _HQQ_REGISTRY

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def _make_quantizable(cls, model, quantized: bool):
        model.hqq_quantized = quantized
        model.arch_key = model.default_cfg["architecture"]
        model.quantize_model = (
            lambda quant_config,
            compute_dtype=torch.float16,
            device="cuda": cls.quantize_model_(
                model=model,
                quant_config=quant_config,
                compute_dtype=compute_dtype,
                device=device,
            )
        )
        model.save_quantized = lambda save_dir: cls.save_quantized_(
            model=model, save_dir=save_dir
        )
        model.cuda = lambda *args, **kwargs: model if (quantized) else model.cuda
        model.to = lambda *args, **kwargs: model if (quantized) else model.to
        model.float = lambda *args, **kwargs: model if (quantized) else model.float
        model.half = lambda *args, **kwargs: model if (quantized) else model.half
        model.base_class = cls._get_hqq_class(model)

    @classmethod
    def _validate_params(cls, params: dict):
        pass

    @classmethod
    def create_model(cls, *args, **kwargs):
        cls._validate_params(kwargs)
        model = timm.create_model(*args, **kwargs)
        cls._make_quantizable(model, quantized=False)
        return model

    @classmethod
    def _get_arch_key_from_save_dir(cls, save_dir: str):
        with open(BaseHQQModel.get_config_file(save_dir), "r") as file:
            config = json.load(file)
        return config["architecture"]



from setuptools import setup
from torch.utils.cpp_extension import CUDAExtension, BuildExtension

setup(
    name="hqq_aten",
    ext_modules=[
        CUDAExtension(
            "hqq_aten",
            [
                "hqq_aten_cuda.cpp",
                "hqq_aten_cuda_kernel.cu",
            ],
        )
    ],
    extra_compile_args={"cxx": ["-O3"], "nvcc": ["-O3", "--use_fast_math"]},
    cmdclass={"build_ext": BuildExtension},
)

# python3 setup_cuda.py install



# Written by Dr. Hicham Badri @Mobius Labs GmbH - 2023
#####################################################
import torch
import gc
import numpy as np
from typing import Union


def cleanup() -> None:
    try:
        torch.cuda.empty_cache()
    except Exception:
        pass
    gc.collect()


def is_divisible(val1: int, val2: int) -> bool:
    return int(val2 * np.ceil(val1 / val2)) == val1


def zero_pad_row(
    tensor: torch.Tensor, num_rows: int, dtype: Union[torch.dtype, None] = None
) -> torch.Tensor:
    out = torch.zeros(
        [num_rows, tensor.shape[1]],
        device=tensor.device,
        dtype=tensor.dtype if (dtype is None) else dtype,
    )
    out[: len(tensor)] = tensor

    return out



# Written by Dr. Hicham Badri @Mobius Labs GmbH - 2023
#####################################################

from setuptools import setup, find_packages
from setuptools.command.install import install
from setuptools.command.develop import develop
from setuptools.command.egg_info import egg_info
import os


def install_cuda_cmd() -> str:
    cmd = "cd hqq/kernels; "
    cmd += "python setup_cuda.py install; "
    cmd += "cd ../..;"
    return cmd


def run_setup_cuda():
    print("Running setup_cuda.py...")
    try:
        os.system(install_cuda_cmd())
    except Exception as e:
        print("Error while running setup_cuda.py:", e)


class InstallCommand(install):
    def run(self):
        install.run(self)
        run_setup_cuda()


class DevelopCommand(develop):
    def run(self):
        develop.run(self)
        run_setup_cuda()


class EgginfoCommand(egg_info):
    def run(self):
        egg_info.run(self)
        run_setup_cuda()


setup(
    name="hqq",
    version="0.1.7.post3",
    description="Half-Quadratic Quantization (HQQ)",
    url="https://github.com/mobiusml/hqq/",
    author="Dr. Hicham Badri",
    author_email="hicham@mobiuslabs.com",
    license="Apache 2",
    packages=find_packages(include=["hqq", "hqq.*"]),
    package_data={
        "hqq": ["kernels/*.cpp", "kernels/*.cu"],
    },
    include_package_data=True,
    cmdclass={
        "install": InstallCommand,
        "develop": DevelopCommand,
        "egg_info": EgginfoCommand,
    },
    install_requires=[
        "numpy>=1.24.4",
        "tqdm>=4.64.1",
        "einops",
        "accelerate",
        "transformers>=4.36.1",
        "huggingface_hub",
        "termcolor",
        #"timm",
    ],
)
